# pylint:disable=unused-import
from __future__ import annotations
import logging
from collections import defaultdict
from collections.abc import Iterable
from typing import Any, TYPE_CHECKING

import networkx
from cle import SymbolType

from angr import ailment
from angr.analyses.cfg import CFGFast
from angr.knowledge_plugins.functions.function import Function
from angr.knowledge_base import KnowledgeBase
from angr.sim_variable import SimMemoryVariable, SimRegisterVariable, SimStackVariable
from angr.utils import timethis
from angr.analyses import Analysis, AnalysesHub
from .clinic import ClinicStage
from .structured_codegen.c import CStructuredCodeGenerator
from .structuring import RecursiveStructurer, PhoenixStructurer, DEFAULT_STRUCTURER
from .region_identifier import RegionIdentifier
from .optimization_passes.optimization_pass import OptimizationPassStage
from .ailgraph_walker import AILGraphWalker
from .condition_processor import ConditionProcessor
from .decompilation_options import DecompilationOption, PARAM_TO_OPTION
from .decompilation_cache import DecompilationCache
from .utils import remove_edges_in_ailgraph
from .sequence_walker import SequenceWalker
from .structuring.structurer_nodes import SequenceNode
from .presets import DECOMPILATION_PRESETS, DecompilationPreset
from .notes import DecompilationNote

if TYPE_CHECKING:
    from angr.knowledge_plugins.cfg.cfg_model import CFGModel
    from .peephole_optimizations import PeepholeOptimizationExprBase, PeepholeOptimizationStmtBase
    from angr.analyses.typehoon.typevars import TypeVariable, TypeConstraint

l = logging.getLogger(name=__name__)

_PEEPHOLE_OPTIMIZATIONS_TYPE = (
    Iterable[type["PeepholeOptimizationStmtBase"] | type["PeepholeOptimizationExprBase"]] | None
)


class Decompiler(Analysis):
    """
    The decompiler analysis.

    Run this on a Function object for which a normalized CFG has been constructed.
    The fully processed output can be found in result.codegen.text
    """

    def __init__(
        self,
        func: Function | str | int,
        cfg: CFGFast | CFGModel | None = None,
        options=None,
        preset: str | DecompilationPreset | None = None,
        optimization_passes=None,
        sp_tracker_track_memory=True,
        variable_kb=None,
        peephole_optimizations: _PEEPHOLE_OPTIMIZATIONS_TYPE = None,
        vars_must_struct: set[str] | None = None,
        flavor="pseudocode",
        expr_comments=None,
        stmt_comments=None,
        ite_exprs=None,
        binop_operators=None,
        decompile=True,
        regen_clinic=True,
        inline_functions=None,
        desired_variables=None,
        update_memory_data: bool = True,
        want_full_graph: bool = False,
        generate_code: bool = True,
        use_cache: bool = True,
        expr_collapse_depth: int = 16,
        clinic_graph=None,
        clinic_arg_vvars=None,
        clinic_start_stage=None,
        clinic_end_stage=None,
        clinic_skip_stages=(),
        static_vvars: dict | None = None,
        static_buffers: dict | None = None,
    ):
        if not isinstance(func, Function):
            func = self.kb.functions[func]
        self.func: Function = func

        if cfg is None:
            cfg = self.func._function_manager._kb.cfgs.get_most_accurate()
        self._cfg = cfg.model if isinstance(cfg, CFGFast) else cfg
        self._options = self._parse_options(options) if options else []

        if preset is None and optimization_passes:
            self._optimization_passes = optimization_passes
        else:
            # we use the preset
            if isinstance(preset, str):
                if preset not in DECOMPILATION_PRESETS:
                    raise KeyError(f"Decompilation preset {preset} is not found")
                preset = DECOMPILATION_PRESETS[preset]
            elif preset is None:
                preset = DECOMPILATION_PRESETS["default"]
            if not isinstance(preset, DecompilationPreset):
                raise TypeError('"preset" must be a DecompilationPreset instance')
            self._optimization_passes = preset.get_optimization_passes(self.project.arch, self.project.simos.name)
        l.debug("Get %d optimization passes for the current binary.", len(self._optimization_passes))
        self._sp_tracker_track_memory = sp_tracker_track_memory
        self._peephole_optimizations = peephole_optimizations
        self._vars_must_struct = vars_must_struct
        self._flavor = flavor
        self._variable_kb = variable_kb
        self._expr_comments = expr_comments
        self._stmt_comments = stmt_comments
        self._ite_exprs = ite_exprs
        self._binop_operators = binop_operators
        self._regen_clinic = regen_clinic
        self._update_memory_data = update_memory_data
        self._want_full_graph = want_full_graph
        self._generate_code = generate_code
        self._inline_functions = frozenset(inline_functions) if inline_functions else set()
        self._desired_variables = frozenset(desired_variables) if desired_variables else set()
        self._static_vvars = static_vvars if static_vvars is not None else {}
        self._static_buffers = static_buffers if static_buffers is not None else {}
        self._cache_parameters = (
            {
                "cfg": self._cfg,
                "variable_kb": self._variable_kb,
                "options": {(o, v) for o, v in self._options if o.category != "Display" and v != o.default_value},
                "optimization_passes": self._optimization_passes,
                "sp_tracker_track_memory": self._sp_tracker_track_memory,
                "peephole_optimizations": self._peephole_optimizations,
                "vars_must_struct": self._vars_must_struct,
                "flavor": self._flavor,
                "expr_comments": self._expr_comments,
                "stmt_comments": self._stmt_comments,
                "ite_exprs": self._ite_exprs,
                "binop_operators": self._binop_operators,
                "inline_functions": self._inline_functions,
                "desired_variables": self._desired_variables,
                "static_vvars": self._static_vvars,
                "static_buffers": self._static_buffers,
            }
            if use_cache
            else None
        )

        self.clinic = None  # mostly for debugging purposes
        self._clinic_graph = clinic_graph
        self._clinic_arg_vvars = clinic_arg_vvars
        self._clinic_start_stage = clinic_start_stage
        self._clinic_end_stage = clinic_end_stage
        self._clinic_skip_stages = clinic_skip_stages
        self.codegen: CStructuredCodeGenerator | None = None
        self.cache: DecompilationCache | None = None
        self.options_by_class = None
        self.seq_node: SequenceNode | None = None
        self.unoptimized_ail_graph: networkx.DiGraph | None = None
        self.ail_graph: networkx.DiGraph | None = None
        self.vvar_id_start = None
        self._copied_var_ids: set[int] = set()
        self._optimization_scratch: dict[str, Any] = {}
        self.expr_collapse_depth = expr_collapse_depth
        self.notes: dict[str, DecompilationNote] = {}
        self.region_identifier = None

        if decompile:
            with self._resilience():
                self._decompile()
            if self.errors:
                if (self.func.addr, self._flavor) not in self.kb.decompilations:
                    self.kb.decompilations[(self.func.addr, self._flavor)] = DecompilationCache(self.func.addr)
                for error in self.errors:
                    self.kb.decompilations[(self.func.addr, self._flavor)].errors.append(error.format())
                with self._resilience():
                    l.info("Decompilation failed for %s. Switching to basic preset and trying again.", self.func)
                    if preset != DECOMPILATION_PRESETS["basic"]:
                        self._optimization_passes = DECOMPILATION_PRESETS["basic"].get_optimization_passes(
                            self.project.arch, self.project.simos.name
                        )
                        self._decompile()
                        for error in self.errors:
                            self.kb.decompilations[(self.func.addr, self._flavor)].errors.append(error.format())

    def _can_use_decompilation_cache(self, cache: DecompilationCache) -> bool:
        if self._cache_parameters is None or cache.parameters is None:
            return False
        a, b = self._cache_parameters, cache.parameters
        id_checks = {"cfg", "variable_kb"}
        return all(a[k] is b[k] if k in id_checks else a[k] == b[k] for k in self._cache_parameters)

    @staticmethod
    def _parse_options(options: list[tuple[DecompilationOption | str, Any]]) -> list[tuple[DecompilationOption, Any]]:
        """
        Parse the options and return a list of option tuples.
        """

        converted_options = []
        for o, v in options:
            if isinstance(o, str):
                # convert to DecompilationOption
                o = PARAM_TO_OPTION[o]
            converted_options.append((o, v))
        return converted_options

    @timethis
    def _decompile(self):
        if self.func.is_simprocedure:
            return

        cache = None

        if self._cache_parameters is not None:
            try:
                cache = self.kb.decompilations[(self.func.addr, self._flavor)]
                if not self._can_use_decompilation_cache(cache):
                    cache = None
            except KeyError:
                pass

        if cache:
            old_codegen = cache.codegen
            old_clinic = cache.clinic
            ite_exprs = cache.ite_exprs if self._ite_exprs is None else self._ite_exprs
            binop_operators = cache.binop_operators if self._binop_operators is None else self._binop_operators
            l.debug("Decompilation cache hit")
        else:
            old_codegen = None
            old_clinic = None
            ite_exprs = self._ite_exprs
            binop_operators = self._binop_operators
            l.debug("Decompilation cache miss")

        self.options_by_class = defaultdict(list)

        if self._options:
            for o, v in self._options:
                self.options_by_class[o.cls].append((o, v))

        # set global variables
        self._set_global_variables()
        self._update_progress(5.0, text="Converting to AIL")

        variable_kb = self._variable_kb
        # fall back to old codegen
        if variable_kb is None and old_codegen is not None and isinstance(old_codegen, CStructuredCodeGenerator):
            variable_kb = old_codegen._variable_kb

        if variable_kb is None:
            reset_variable_names = True
        else:
            reset_variable_names = self.func.addr not in variable_kb.variables.function_managers

        # determine a few arguments according to the structuring algorithm
        fold_callexprs_into_conditions = False
        self._force_loop_single_exit = True
        self._refine_loops_with_single_successor = False
        self._complete_successors = False
        self._recursive_structurer_params = self.options_to_params(self.options_by_class["recursive_structurer"])
        if "structurer_cls" not in self._recursive_structurer_params:
            self._recursive_structurer_params["structurer_cls"] = DEFAULT_STRUCTURER
        # is the algorithm based on Phoenix (a schema-based algorithm)?
        if issubclass(self._recursive_structurer_params["structurer_cls"], PhoenixStructurer):
            self._force_loop_single_exit = False
            # self._refine_loops_with_single_successor = True
            self._complete_successors = True
            fold_callexprs_into_conditions = True

        cache = DecompilationCache(self.func.addr)
        if self._cache_parameters is not None:
            cache.parameters = self._cache_parameters
        cache.ite_exprs = ite_exprs
        cache.binop_operators = binop_operators

        # convert function blocks to AIL blocks
        def progress_callback(p, **kwargs):
            return self._update_progress(p * (70 - 5) / 100.0 + 5, **kwargs)

        if self._regen_clinic or old_clinic is None or self.func.prototype is None:
            clinic = self.project.analyses.Clinic(
                self.func,
                kb=self.kb,
                fail_fast=self._fail_fast,
                variable_kb=variable_kb,
                reset_variable_names=reset_variable_names,
                optimization_passes=self._optimization_passes,
                sp_tracker_track_memory=self._sp_tracker_track_memory,
                fold_callexprs_into_conditions=fold_callexprs_into_conditions,
                cfg=self._cfg,
                peephole_optimizations=self._peephole_optimizations,
                must_struct=self._vars_must_struct,
                cache=cache,
                progress_callback=progress_callback,
                inline_functions=self._inline_functions,
                desired_variables=self._desired_variables,
                optimization_scratch=self._optimization_scratch,
                force_loop_single_exit=self._force_loop_single_exit,
                refine_loops_with_single_successor=self._refine_loops_with_single_successor,
                complete_successors=self._complete_successors,
                ail_graph=self._clinic_graph,
                arg_vvars=self._clinic_arg_vvars,
                start_stage=self._clinic_start_stage,
                end_stage=self._clinic_end_stage,
                skip_stages=self._clinic_skip_stages,
                notes=self.notes,
                static_vvars=self._static_vvars,
                static_buffers=self._static_buffers,
                **self.options_to_params(self.options_by_class["clinic"]),
            )
        else:
            clinic = old_clinic
            # reuse the old, unaltered graph
            clinic.graph = clinic.cc_graph
            clinic.cc_graph = clinic.copy_graph()

        self.clinic = clinic
        self.cache = cache
        self._variable_kb = clinic.variable_kb
        self._update_progress(70.0, text="Identifying regions")
        self.vvar_id_start = clinic.vvar_id_start
        self._copied_var_ids = clinic.copied_var_ids

        if clinic.graph is None:
            # the function is empty
            return

        # expose a copy of the graph before any optimizations that may change the graph occur;
        # use this graph if you need a reference of exact mapping of instructions to AIL statements
        self.unoptimized_ail_graph = (
            clinic.unoptimized_graph if clinic.unoptimized_graph is not None else clinic.copy_graph()
        )
        cond_proc = ConditionProcessor(self.project.arch)

        clinic.graph = self._run_graph_simplification_passes(
            clinic.graph,
            clinic.reaching_definitions,
            ite_exprs=ite_exprs,
        )

        # recover regions, delay updating when we have optimizations that may update regions themselves
        delay_graph_updates = any(
            pass_.STAGE == OptimizationPassStage.DURING_REGION_IDENTIFICATION for pass_ in self._optimization_passes
        )
        self.region_identifier = self._recover_regions(clinic.graph, cond_proc, update_graph=not delay_graph_updates)

        self._update_progress(73.0, text="Running region-simplification passes")

        # run optimizations that may require re-RegionIdentification
        clinic.graph, self.region_identifier = self._run_region_simplification_passes(
            clinic.graph,
            self.region_identifier,
            clinic.reaching_definitions,
            ite_exprs=ite_exprs,
            arg_vvars=set(clinic.arg_vvars),
            edges_to_remove=clinic.edges_to_remove,
        )

        if not self._want_full_graph:
            # finally (no more graph-based simplifications will run in the future),
            # we can remove the edges that should be removed!
            remove_edges_in_ailgraph(clinic.graph, clinic.edges_to_remove)

        # save the graph before structuring happens (for AIL view)
        clinic.cc_graph = clinic.copy_graph()

        codegen = None
        seq_node = None
        # in the event that the decompiler is used without code generation as the target, we should avoid all
        # heavy analysis that is used only for the purpose of code generation
        # we also do not want to run structurer if clinic stopped before variable recovery
        if self._generate_code and (
            self._clinic_end_stage is None or self._clinic_end_stage >= ClinicStage.RECOVER_VARIABLES
        ):
            self._update_progress(75.0, text="Structuring code")

            # structure it
            rs = self.project.analyses[RecursiveStructurer].prep(kb=self.kb, fail_fast=self._fail_fast)(
                self.region_identifier.region,
                cond_proc=cond_proc,
                func=self.func,
                **self._recursive_structurer_params,
            )
            self._update_progress(80.0, text="Simplifying regions")

            # simplify it
            s = self.project.analyses.RegionSimplifier(
                self.func,
                rs.result,
                arg_vvars=set(self.clinic.arg_vvars),
                kb=self.kb,
                fail_fast=self._fail_fast,
                **self.options_to_params(self.options_by_class["region_simplifier"]),
            )
            seq_node = s.result
            seq_node = self._run_post_structuring_simplification_passes(
                seq_node, binop_operators=cache.binop_operators, goto_manager=s.goto_manager, graph=clinic.graph
            )

            # rewrite the sequence node to remove phi expressions
            seq_node = self.transform_seqnode_from_ssa(seq_node)

            # update memory data
            if self._cfg is not None and self._update_memory_data:
                self.find_data_references_and_update_memory_data(seq_node)

            if self._clinic_end_stage is None or self._clinic_end_stage >= ClinicStage.RECOVER_VARIABLES:
                self._update_progress(85.0, text="Generating code")
                codegen = self.project.analyses.StructuredCodeGenerator(
                    self.func,
                    seq_node,
                    cfg=self._cfg,
                    ail_graph=clinic.graph,
                    flavor=self._flavor,
                    func_args=clinic.arg_list,
                    kb=self.kb,
                    fail_fast=self._fail_fast,
                    variable_kb=clinic.variable_kb,
                    expr_comments=old_codegen.expr_comments if old_codegen is not None else None,
                    stmt_comments=old_codegen.stmt_comments if old_codegen is not None else None,
                    const_formats=old_codegen.const_formats if old_codegen is not None else None,
                    externs=clinic.externs,
                    binop_depth_cutoff=self.expr_collapse_depth,
                    notes=self.notes,
                    **self.options_to_params(self.options_by_class["codegen"]),
                )

        self._update_progress(90.0, text="Finishing up")
        self.seq_node = seq_node
        self.codegen = codegen
        # save a copy of the AIL graph that is optimized but not modified by region identification
        self.ail_graph = clinic.cc_graph
        self.cache.codegen = codegen
        self.cache.clinic = self.clinic

        self.kb.decompilations[(self.func.addr, self._flavor)] = self.cache
        self._finish_progress()

    def _recover_regions(self, graph: networkx.DiGraph, condition_processor, update_graph: bool = True):
        return self.project.analyses[RegionIdentifier].prep(kb=self.kb, fail_fast=self._fail_fast)(
            self.func,
            graph=graph,
            cond_proc=condition_processor,
            update_graph=update_graph,
            force_loop_single_exit=self._force_loop_single_exit,
            refine_loops_with_single_successor=self._refine_loops_with_single_successor,
            complete_successors=self._complete_successors,
            entry_node_addr=self.clinic.entry_node_addr,
            **self.options_to_params(self.options_by_class["region_identifier"]),
        )

    @timethis
    def _run_graph_simplification_passes(self, ail_graph, reaching_definitions, **kwargs):
        """
        Runs optimizations that should be executed before region identification.

        :param ail_graph:   DiGraph with AIL Statements
        :param reaching_definitions: ReachingDefinitionAnalysis
        :return:            The possibly new AIL DiGraph and RegionIdentifier
        """
        addr_and_idx_to_blocks: dict[tuple[int, int | None], ailment.Block] = {}
        addr_to_blocks: dict[int, set[ailment.Block]] = defaultdict(set)

        # update blocks_map to allow node_addr to node lookup
        def _updatedict_handler(node):
            addr_and_idx_to_blocks[(node.addr, node.idx)] = node
            addr_to_blocks[node.addr].add(node)

        AILGraphWalker(ail_graph, _updatedict_handler).walk()

        # run each pass
        for pass_ in self._optimization_passes:
            # only for post region id opts
            if pass_.STAGE != OptimizationPassStage.BEFORE_REGION_IDENTIFICATION:
                continue
            if pass_.STRUCTURING and self._recursive_structurer_params["structurer_cls"].NAME not in pass_.STRUCTURING:
                l.warning(
                    "Skipping %s because it does not support structuring algorithm: %s",
                    pass_,
                    self._recursive_structurer_params["structurer_cls"].NAME,
                )
                continue

            pass_ = timethis(pass_)
            a = pass_(
                self.func,
                blocks_by_addr=addr_to_blocks,
                blocks_by_addr_and_idx=addr_and_idx_to_blocks,
                graph=ail_graph,
                variable_kb=self._variable_kb,
                reaching_definitions=reaching_definitions,
                entry_node_addr=self.clinic.entry_node_addr,
                scratch=self._optimization_scratch,
                force_loop_single_exit=self._force_loop_single_exit,
                refine_loops_with_single_successor=self._refine_loops_with_single_successor,
                complete_successors=self._complete_successors,
                **kwargs,
            )

            # should be None if no changes
            if a.out_graph:
                # use the new graph
                ail_graph = a.out_graph

        return ail_graph

    @timethis
    def _run_region_simplification_passes(self, ail_graph, ri, reaching_definitions, arg_vvars: set[int], **kwargs):
        """
        Runs optimizations that should be executed after a single region identification. This function will return
        two items: the new RegionIdentifier object and the new AIL Graph, which should probably be written
        back to the clinic object that the graph is from.

        Note: After each optimization run, if the optimization modifies the graph in any way then RegionIdentification
        will be run again.

        :param ail_graph:   DiGraph with AIL Statements
        :param ri:          RegionIdentifier
        :param reaching_definitions: ReachingDefinitionAnalysis
        :return:            The possibly new AIL DiGraph and RegionIdentifier
        """
        addr_and_idx_to_blocks: dict[tuple[int, int | None], ailment.Block] = {}
        addr_to_blocks: dict[int, set[ailment.Block]] = defaultdict(set)

        # update blocks_map to allow node_addr to node lookup
        def _updatedict_handler(node):
            addr_and_idx_to_blocks[(node.addr, node.idx)] = node
            addr_to_blocks[node.addr].add(node)

        AILGraphWalker(ail_graph, _updatedict_handler).walk()

        # run each pass
        for pass_ in self._optimization_passes:
            # only for post region id opts
            if pass_.STAGE != OptimizationPassStage.DURING_REGION_IDENTIFICATION:
                continue
            if pass_.STRUCTURING and self._recursive_structurer_params["structurer_cls"].NAME not in pass_.STRUCTURING:
                l.warning(
                    "Skipping %s because it does not support structuring algorithm: %s",
                    pass_,
                    self._recursive_structurer_params["structurer_cls"].NAME,
                )
                continue

            pass_ = timethis(pass_)
            a = pass_(
                self.func,
                blocks_by_addr=addr_to_blocks,
                blocks_by_addr_and_idx=addr_and_idx_to_blocks,
                graph=ail_graph,
                variable_kb=self._variable_kb,
                arg_vvars=arg_vvars,
                region_identifier=ri,
                reaching_definitions=reaching_definitions,
                vvar_id_start=self.vvar_id_start,
                entry_node_addr=self.clinic.entry_node_addr,
                scratch=self._optimization_scratch,
                force_loop_single_exit=self._force_loop_single_exit,
                refine_loops_with_single_successor=self._refine_loops_with_single_successor,
                complete_successors=self._complete_successors,
                peephole_optimizations=self._peephole_optimizations,
                avoid_vvar_ids=self._copied_var_ids,
                **kwargs,
            )

            # should be None if no changes
            if a.out_graph:
                # use the new graph
                ail_graph = a.out_graph

                # the graph might change! update them.
                addr_and_idx_to_blocks = {}
                addr_to_blocks = defaultdict(set)
                AILGraphWalker(ail_graph, _updatedict_handler).walk()

                cond_proc = ConditionProcessor(self.project.arch)
                # always update RI on graph change
                ri = self._recover_regions(ail_graph, cond_proc, update_graph=False)

                self.vvar_id_start = a.vvar_id_start

        return ail_graph, self._recover_regions(ail_graph, ConditionProcessor(self.project.arch), update_graph=True)

    @timethis
    def _run_post_structuring_simplification_passes(self, seq_node, **kwargs):
        for pass_ in self._optimization_passes:
            if pass_.STAGE != OptimizationPassStage.AFTER_STRUCTURING:
                continue

            pass_ = timethis(pass_)
            a = pass_(
                self.func,
                seq=seq_node,
                scratch=self._optimization_scratch,
                peephole_optimizations=self._peephole_optimizations,
                **kwargs,
            )
            if a.out_seq:
                seq_node = a.out_seq

        return seq_node

    def _set_global_variables(self):
        global_variables = self.kb.variables["global"]
        for symbol in self.project.loader.main_object.symbols:
            if symbol.type == SymbolType.TYPE_OBJECT:
                ident = global_variables.next_variable_ident("global")
                global_variables.set_variable(
                    "global",
                    symbol.rebased_addr,
                    SimMemoryVariable(symbol.rebased_addr, 1, name=symbol.name, ident=ident),
                )

    def reflow_variable_types(self, cache: DecompilationCache):
        """
        Re-run type inference on an existing variable recovery result, then rerun codegen to generate new results.

        :return:
        """

        # extract everything from the cache
        type_constraints: dict[TypeVariable, set[TypeConstraint]] = cache.type_constraints
        func_typevar = cache.func_typevar
        var_to_typevar = cache.var_to_typevar
        arg_vvars = cache.arg_vvars
        stack_offset_typevars = cache.stack_offset_typevars
        stackvar_max_sizes = cache.stackvar_max_sizes
        codegen = cache.codegen

        var_kb = self._variable_kb if self._variable_kb is not None else KnowledgeBase(self.project)

        if self.func.addr not in var_kb.variables:
            # for some reason variables for the current function don't really exist...
            groundtruth = {}
        else:
            var_manager = var_kb.variables[self.func.addr]
            # ground-truth types
            groundtruth = {}
            for variable in var_manager.variables_with_manual_types:
                vartype = var_manager.variable_to_types.get(variable, None)
                if vartype is not None:
                    for typevar in var_to_typevar[variable]:
                        groundtruth[typevar] = vartype

        if self.func.prototype is not None and not self.func.is_prototype_guessed:
            for arg_i, (_, variable) in arg_vvars.items():
                if arg_i < len(self.func.prototype.args):
                    for tv in var_to_typevar[variable]:
                        groundtruth[tv] = self.func.prototype.args[arg_i]

        # variables that must be interpreted as structs
        if self._vars_must_struct:
            must_struct = set()
            for var, typevars in var_to_typevar.items():
                for typevar in typevars:
                    if var.ident in self._vars_must_struct:
                        must_struct.add(typevar)
        else:
            must_struct = None

        tv_max_sizes = {}
        for v, s in stackvar_max_sizes.items():
            assert isinstance(v, SimStackVariable)
            if v in var_to_typevar:
                for tv in var_to_typevar[v]:
                    tv_max_sizes[tv] = s
            if v.offset in stack_offset_typevars:
                tv = stack_offset_typevars[v.offset]
                tv_max_sizes[tv] = s

        # Type inference
        try:
            tp = self.project.analyses.Typehoon(
                type_constraints,
                func_typevar,
                kb=var_kb,
                fail_fast=self._fail_fast,
                var_mapping=var_to_typevar,
                must_struct=must_struct,
                ground_truth=groundtruth,
                stack_offset_tvs=stack_offset_typevars,
                stackvar_max_sizes=tv_max_sizes,
            )
            tp.update_variable_types(
                self.func.addr,
                {v: t for v, t in var_to_typevar.items() if isinstance(v, (SimRegisterVariable, SimStackVariable))},
            )
            tp.update_variable_types(
                "global",
                {v: t for v, t in var_to_typevar.items() if isinstance(v, (SimRegisterVariable, SimStackVariable))},
            )
            # update the function prototype if needed
            if self.func.is_prototype_guessed and self.func.prototype is not None and self.func.prototype.args:
                var_manager = var_kb.variables[self.func.addr]
                for i, arg in enumerate(codegen.cfunc.arg_list):
                    if i >= len(self.func.prototype.args):
                        break
                    var = arg.variable
                    new_type = var_manager.get_variable_type(var)
                    if new_type is not None:
                        self.func.prototype.args = (
                            *self.func.prototype.args[:i],
                            new_type,
                            *self.func.prototype.args[i + 1 :],
                        )
        except Exception:  # pylint:disable=broad-except
            if self._fail_fast:
                raise
            l.warning(
                "Typehoon analysis failed. Variables will not have types. Please report to GitHub.", exc_info=True
            )

        codegen.reload_variable_types()

        return codegen

    def find_data_references_and_update_memory_data(self, seq_node: SequenceNode):
        const_values: set[int] = set()

        def _handle_Const(expr_idx: int, expr: ailment.Expr.Const, *args, **kwargs):  # pylint:disable=unused-argument
            const_values.add(expr.value)

        def _handle_block(block: ailment.Block, **kwargs):  # pylint:disable=unused-argument
            block_walker = ailment.AILBlockViewer(
                expr_handlers={
                    ailment.Expr.Const: _handle_Const,
                }
            )
            block_walker.walk(block)

        seq_walker = SequenceWalker(
            handlers={
                ailment.Block: _handle_block,
            },
            update_seqnode_in_place=False,
        )
        seq_walker.walk(seq_node)

        added_memory_data_addrs = []
        for data_addr in const_values:
            if data_addr in self._cfg.memory_data:
                continue
            if not self.project.loader.find_loadable_containing(data_addr):
                continue
            if self._cfg.add_memory_data(data_addr, None):
                added_memory_data_addrs.append(data_addr)

        self._cfg.tidy_data_references(
            memory_data_addrs=added_memory_data_addrs,
        )

    def transform_graph_from_ssa(self, ail_graph: networkx.DiGraph) -> networkx.DiGraph:
        """
        Translate an SSA AIL graph out of SSA form. This is useful for producing a non-SSA AIL graph for displaying in
        angr management.

        :param ail_graph:   The AIL graph to transform out of SSA form.
        :return:            The translated AIL graph.
        """
        variable_kb = self._variable_kb
        dephication = self.project.analyses.GraphDephication(
            self.func, ail_graph, rewrite=True, variable_kb=variable_kb, kb=self.kb, fail_fast=self._fail_fast
        )
        return dephication.output

    def transform_seqnode_from_ssa(self, seq_node: SequenceNode) -> SequenceNode:
        variable_kb = self._variable_kb
        dephication = self.project.analyses.SeqNodeDephication(
            self.func, seq_node, rewrite=True, variable_kb=variable_kb, kb=self.kb, fail_fast=self._fail_fast
        )
        return dephication.output

    @staticmethod
    def options_to_params(options: list[tuple[DecompilationOption, Any]]) -> dict[str, Any]:
        """
        Convert decompilation options to a dict of params.

        :param options:   The decompilation options.
        :return:          A dict of keyword arguments.
        """

        d = {}
        for option, value in options:
            if option.convert is not None:
                d[option.param] = option.convert(value)
            else:
                d[option.param] = value
        return d


AnalysesHub.register_default("Decompiler", Decompiler)



# pylint:disable=arguments-differ,too-many-boolean-expressions,no-self-use
from __future__ import annotations

from archinfo import Endness
from angr.ailment.constant import UNDETERMINED_SIZE
from angr.ailment.expression import Const, VirtualVariable, BinaryOp, UnaryOp, Load
from angr.ailment.statement import Call, WeakAssignment

from angr.sim_type import SimTypeReference, SimCppClass
from angr.knowledge_plugins.key_definitions import atoms
from .base import PeepholeOptimizationStmtBase


class RewriteCxxOperatorCalls(PeepholeOptimizationStmtBase):
    """
    Rewrite C++ operator function calls into operations.
    """

    __slots__ = ()

    NAME = "Rewrite C++ operator function calls into operations"
    stmt_classes = (Call,)

    def optimize(self, stmt: Call, block=None, **kwargs):  # type: ignore
        assert self.project is not None

        # are we calling a function that we deem as an overridden operator function?
        if isinstance(stmt.target, Const):
            func_addr = stmt.target.value
            if not self.project.kb.functions.contains_addr(func_addr):
                return None
            func = self.project.kb.functions[func_addr]
            if "operator=" in func.demangled_name and stmt.args is not None:
                return self._optimize_operator_equal(stmt)
            if "operator+" in func.demangled_name and stmt.args is not None:
                return self._optimize_operator_add(stmt)
            # TODO: Support other types of C++ operator functions

        return None

    def _optimize_operator_equal(self, stmt: Call) -> WeakAssignment | None:
        if stmt.args and len(stmt.args) == 2 and isinstance(stmt.args[0], UnaryOp) and stmt.args[0].op == "Reference":
            dst = stmt.args[0].operand
            if isinstance(dst, VirtualVariable):
                self.preserve_vvar_ids.add(dst.varid)
                atom = atoms.VirtualVariable(dst.varid, dst.size, dst.category, dst.oident)
                if stmt.prototype is not None and isinstance(stmt.prototype.returnty, SimTypeReference):
                    type_hint = self._type_hint_from_typeref(stmt.prototype.returnty)
                    if type_hint is not None:
                        self.type_hints.append((atom, type_hint))
            arg1 = (
                Load(None, stmt.args[1], UNDETERMINED_SIZE, Endness.BE, **stmt.tags)
                if isinstance(stmt.args[1], Const)
                else stmt.args[1]
            )
            type_ = None
            if stmt.prototype is not None:
                dst_ty = stmt.prototype.returnty
                if isinstance(dst_ty, SimTypeReference):
                    dst_ty = dst_ty.refs
                type_ = {"dst": dst_ty, "src": stmt.prototype.args[1]}
            return WeakAssignment(stmt.idx, stmt.args[0].operand, arg1, type=type_, **stmt.tags)  # type:ignore
        return None

    def _optimize_operator_add(self, stmt: Call) -> WeakAssignment | None:
        if (
            stmt.args
            and len(stmt.args) == 3
            and isinstance(stmt.args[1], UnaryOp)
            and stmt.args[1].op == "Reference"
            and isinstance(stmt.args[1].operand, VirtualVariable)
            and isinstance(stmt.args[2], Const)
            and isinstance(stmt.ret_expr, VirtualVariable)
        ):
            arg2 = Load(None, stmt.args[2], UNDETERMINED_SIZE, Endness.BE, **stmt.tags)
            addition = BinaryOp(None, "Add", [stmt.args[1].operand, arg2], **stmt.tags)
            type_ = None
            if stmt.prototype is not None:
                dst_ty = stmt.prototype.returnty
                if isinstance(dst_ty, SimTypeReference):
                    dst_ty = dst_ty.refs
                type_ = {"dst": dst_ty, "src": stmt.prototype.args[1]}
            return WeakAssignment(stmt.idx, stmt.ret_expr, addition, type=type_, **stmt.tags)
        return None

    @staticmethod
    def _type_hint_from_typeref(typeref: SimTypeReference) -> str | None:
        if isinstance(typeref.refs, SimCppClass) and typeref.refs.unique_name:
            return typeref.refs.unique_name
        return None



from __future__ import annotations

from .base import ExplorationTechnique


class StubStasher(ExplorationTechnique):
    """
    Stash states that reach a stub SimProcedure.
    """

    @staticmethod
    def post_filter(state):
        hook = state.project.hooked_by(state.addr)
        return hook and hook.is_stub

    def step(self, simgr, stash="active", **kwargs):
        simgr.step(stash=stash, **kwargs)
        simgr.move(stash, "stub", filter_func=self.post_filter)
        return simgr



from __future__ import annotations

from pyvex import IRSB
from pyvex.stmt import WrTmp


def get_tmp_def_stmt(vex_block: IRSB, tmp_idx: int) -> int | None:
    for i, stmt in enumerate(vex_block.statements):
        if isinstance(stmt, WrTmp) and stmt.tmp == tmp_idx:
            return i
    return None



from __future__ import annotations
import archinfo
import claripy
from angr.analyses.reaching_definitions.function_handler import FunctionCallDataUnwrapped, FunctionHandler
from angr.analyses.reaching_definitions.rd_state import ReachingDefinitionsState
from angr.knowledge_plugins.key_definitions.live_definitions import DerefSize
from angr.knowledge_plugins.key_definitions.live_definitions import MultiValues

# pylint: disable=no-self-use,missing-class-docstring,unused-argument


class LibcStringHandlers(FunctionHandler):
    @FunctionCallDataUnwrapped.decorate
    def handle_impl_strcat(self, state: ReachingDefinitionsState, data: FunctionCallDataUnwrapped):
        src0_atom = state.deref(data.args_atoms[0], DerefSize.NULL_TERMINATE)
        src1_atom = state.deref(data.args_atoms[1], DerefSize.NULL_TERMINATE)
        src0_value = state.get_values(src0_atom) if src0_atom is not None else None
        src1_value = state.get_values(src1_atom) if src1_atom is not None else None

        if src0_value is not None and src1_value is not None:
            src0_value = src0_value.extract(0, len(src0_value) // 8 - 1, archinfo.Endness.BE)
            dest_value = src0_value.concat(src1_value)
            dest_atom = state.deref(data.args_atoms[0], len(dest_value) // 8, endness=archinfo.Endness.BE)
        elif src0_value is not None:
            src0_value = src0_value.extract(0, len(src0_value) // 8 - 1, archinfo.Endness.BE)
            top_val = state.top(state.arch.bits)
            if src1_atom is not None:
                for defn in state.get_definitions(src1_atom):
                    top_val = state.annotate_with_def(top_val, defn)
            dest_value = src0_value.concat(MultiValues(top_val))
            dest_atom = state.deref(data.args_atoms[0], len(dest_value) // 8, endness=archinfo.Endness.BE)
        else:
            dest_value = state.top(state.arch.bits)
            dest_atom = src0_atom
        if src0_atom is not None and src1_atom is not None:
            data.depends(dest_atom, src0_atom, src1_atom, value=dest_value)
        data.depends(data.ret_atoms, data.args_atoms[0], value=src0_value)

    handle_impl_strncat = handle_impl_strcat

    @FunctionCallDataUnwrapped.decorate
    def handle_impl_strlen(self, state: ReachingDefinitionsState, data: FunctionCallDataUnwrapped):
        src_atom = state.deref(data.args_atoms[0], DerefSize.NULL_TERMINATE)
        if src_atom is not None:
            src_str = state.get_values(src_atom) if src_atom is not None else None
            if src_str is not None:
                data.depends(data.ret_atoms, src_atom, value=len(src_str) // 8 - 1)
            else:
                data.depends(data.ret_atoms, src_atom)
        else:
            data.depends(data.ret_atoms, data.args_atoms[0])

    @FunctionCallDataUnwrapped.decorate
    def handle_impl_strcpy(self, state: ReachingDefinitionsState, data: FunctionCallDataUnwrapped):
        src_atom = state.deref(data.args_atoms[1], DerefSize.NULL_TERMINATE)
        src_str = state.get_values(src_atom) if src_atom is not None else None
        if src_str is None:
            src_str = state.top(state.arch.bits)
            if src_atom is not None:
                for defn in state.get_definitions(src_atom):
                    src_str = state.annotate_with_def(src_str, defn)
            src_str = MultiValues(src_str)

        dst_atom = state.deref(data.args_atoms[0], len(src_str) // 8)
        if src_atom is not None:
            data.depends(dst_atom, src_atom, value=src_str)
        data.depends(data.ret_atoms, data.args_atoms[0], value=state.get_values(data.args_atoms[0]))

    @FunctionCallDataUnwrapped.decorate
    def handle_impl_strncpy(self, state: ReachingDefinitionsState, data: FunctionCallDataUnwrapped):
        n = state.get_concrete_value(data.args_atoms[2])
        src_atom = state.deref(data.args_atoms[1], DerefSize.NULL_TERMINATE)
        src_str = state.get_values(src_atom) if src_atom is not None else None
        if src_str is None and src_atom is not None:
            tmp_atom = state.deref(data.args_atoms[1], 1)
            if tmp_atom is not None:
                tmp_str = state.get_values(tmp_atom)
                val_defns = None if tmp_str is None else state.get_definitions(tmp_str)
                if tmp_str is None or not val_defns:  # There's no data at all or no valid definitions
                    src_str = state.top(state.arch.bits if n is None or n > state.arch.bytes else n * 8)
                    defns = state.get_definitions(src_atom) if src_atom is not None else []
                    for defn in defns:
                        src_str = state.annotate_with_def(src_str, defn)
                    src_str = MultiValues(src_str)
                else:  # We found some data, but it's not NULL_TERIMINATED or of size n
                    src_atoms = set()
                    for defn in val_defns:
                        a = defn.atom
                        a.size = a.size if n is None or a.size < n else n
                        src_atoms.add(a)
                    src_str = state.get_values(src_atoms)

        elif n is not None and src_str is not None and n < len(src_str) // 8:
            # We have a src_str, but need to truncate it if n is not None and less than the size of src_str
            src_atom = state.deref(data.args_atoms[1], n)
            if src_atom is not None:
                src_str = state.get_values(src_atom)

        if src_str is not None and src_atom is not None:
            dst_atom = state.deref(data.args_atoms[0], len(src_str) // 8)
            data.depends(dst_atom, src_atom, value=src_str)

        data.depends(data.ret_atoms, data.args_atoms[0], value=state.get_values(data.args_atoms[0]))

    @FunctionCallDataUnwrapped.decorate
    def handle_impl_strdup(self, state: ReachingDefinitionsState, data: FunctionCallDataUnwrapped):
        src_atom = state.deref(data.args_atoms[0], DerefSize.NULL_TERMINATE)
        if src_atom is not None:
            src_str = state.get_values(src_atom)
            malloc_size = len(src_str) // 8 if src_str is not None else 1
            heap_ptr = state.heap_allocator.allocate(malloc_size)
            heap_addr = state.heap_address(heap_ptr)
            dst_atom = state.deref(heap_addr, malloc_size)
            data.depends(dst_atom, src_atom, value=src_str)
        data.depends(data.ret_atoms, data.args_atoms[0], value=state.get_values(data.args_atoms[0]))

    @FunctionCallDataUnwrapped.decorate
    def handle_impl_memcpy(self, state: ReachingDefinitionsState, data: FunctionCallDataUnwrapped):
        size = state.get_concrete_value(data.args_atoms[2])
        if size is not None:
            src_atom = state.deref(data.args_atoms[1], size)
            dst_atom = state.deref(data.args_atoms[0], size)
            if src_atom is not None:
                data.depends(dst_atom, src_atom, value=state.get_values(src_atom))
        else:
            # When size is unknown, use default size and create TOP value
            default_size = state.arch.bytes
            src_atom = state.deref(data.args_atoms[1], default_size)
            dst_atom = state.deref(data.args_atoms[0], default_size)
            if src_atom is not None:
                top_val = state.top(default_size * 8)
                for defn in state.get_definitions(src_atom):
                    top_val = state.annotate_with_def(top_val, defn)
                data.depends(dst_atom, src_atom, value=MultiValues(top_val))
        data.depends(data.ret_atoms, data.args_atoms[0], value=state.get_values(data.args_atoms[0]))

    @FunctionCallDataUnwrapped.decorate
    def handle_impl_memset(self, state: ReachingDefinitionsState, data: FunctionCallDataUnwrapped):
        size = state.get_concrete_value(data.args_atoms[2])
        c = state.get_concrete_value(data.args_atoms[1])
        if size is not None:
            dst_atom = state.deref(data.args_atoms[0], size)
            if c is not None:
                value = MultiValues(claripy.BVV(chr(c) * size, size * 8))
                data.depends(dst_atom, data.args_atoms[1], value=value)
            else:
                data.depends(dst_atom, data.args_atoms[1], value=state.get_values(data.args_atoms[1]))

        data.depends(data.ret_atoms, data.args_atoms[0], value=state.get_values(data.args_atoms[0]))

    @FunctionCallDataUnwrapped.decorate
    def handle_impl_strtok(self, state: ReachingDefinitionsState, data: FunctionCallDataUnwrapped):
        # stub: just return the haystack pointer
        data.depends(data.ret_atoms, data.args_atoms[0], value=state.get_values(data.args_atoms[0]))

    handle_impl_strtok_r = handle_impl_strstr = handle_impl_strcasestr = handle_impl_strchr = handle_imple_strrchr = (
        handle_impl_strchrnul
    ) = handle_impl_strtok



from __future__ import annotations

from angr import ailment
from angr.analyses.decompiler.utils import (
    peephole_optimize_expr,
)
from angr.analyses.decompiler.sequence_walker import SequenceWalker
from angr.analyses.decompiler.peephole_optimizations import (
    PeepholeOptimizationExprBase,
    EXPR_OPTS,
)
from .optimization_pass import OptimizationPassStage, SequenceOptimizationPass


class ExpressionSequenceWalker(SequenceWalker):
    """
    Walks sequences with generic expression handling.
    """

    def _handle(self, node, **kwargs):
        if isinstance(node, ailment.Expr.Expression):
            handler = self._handlers.get(ailment.Expr.Expression, None)
            if handler:
                return handler(node, **kwargs)
        return super()._handle(node, **kwargs)


class PostStructuringPeepholeOptimizationPass(SequenceOptimizationPass):
    """
    Perform a post-structuring peephole optimization pass to simplify node statements and expressions.
    """

    ARCHES = None
    PLATFORMS = None
    STAGE = OptimizationPassStage.AFTER_STRUCTURING
    NAME = "Post-Structuring Peephole Optimization"
    DESCRIPTION = (__doc__ or "").strip()

    def __init__(self, func, peephole_optimizations=None, **kwargs):
        super().__init__(func, **kwargs)
        self._peephole_optimizations = peephole_optimizations
        self._expr_peephole_opts = [
            cls(self.project, self.kb, self._func.addr)
            for cls in (self._peephole_optimizations or EXPR_OPTS)
            if issubclass(cls, PeepholeOptimizationExprBase)
        ]
        self.analyze()

    def _check(self):
        return True, None

    def _analyze(self, cache=None):
        walker = ExpressionSequenceWalker(
            handlers={ailment.Expr.Expression: self._optimize_expr, ailment.Block: self._optimize_block}
        )
        walker.walk(self.seq)
        self.out_seq = self.seq

    def _optimize_expr(self, expr, **_):
        new_expr = peephole_optimize_expr(expr, self._expr_peephole_opts)
        return new_expr if expr is not new_expr else None

    def _optimize_block(self, block, **_):
        old_block, new_block = None, block
        while old_block != new_block:
            old_block = new_block
            # Note: AILBlockSimplifier updates expressions in place
            simp = self.project.analyses.AILBlockSimplifier(
                new_block,
                func_addr=self._func.addr,
                peephole_optimizations=self._peephole_optimizations,
            )
            assert simp.result_block is not None
            new_block = simp.result_block
        return new_block if block != new_block else None



# pylint:disable=missing-class-docstring,too-many-boolean-expressions
from __future__ import annotations
from typing import Any
from enum import IntEnum
import string
import logging

import networkx

import claripy

from angr import SIM_LIBRARIES
from angr.calling_conventions import SimRegArg
from angr.errors import SimMemoryMissingError
from angr.knowledge_base import KnowledgeBase
from angr.knowledge_plugins.key_definitions.constants import ObservationPointType
from angr.sim_type import SimTypePointer, SimTypeChar
from angr.analyses import Analysis, AnalysesHub
from angr.procedures.definitions import SimSyscallLibrary
from angr.sim_variable import SimMemoryVariable
from angr.analyses.decompiler.structured_codegen.c import (
    CStructuredCodeWalker,
    CFunctionCall,
    CConstant,
    CAssignment,
    CVariable,
)

from .api_obf_type2_finder import APIObfuscationType2Finder
from .hash_lookup_api_deobfuscator import HashLookupAPIDeobfuscator

_l = logging.getLogger(name=__name__)


class APIObfuscationType(IntEnum):
    TYPE_1 = 0


class APIDeobFuncDescriptor:
    def __init__(self, type_: APIObfuscationType, *, func_addr: int, libname_argidx: int, funcname_argidx: int):
        self.type = type_
        self.func_addr = func_addr
        self.libname_argidx = libname_argidx
        self.funcname_argidx = funcname_argidx


class Type1AssignmentFinder(CStructuredCodeWalker):
    def __init__(self, func_addr: int, desc: APIDeobFuncDescriptor):
        self.func_addr = func_addr
        self.desc = desc
        self.assignments: dict[int, tuple[str, str]] = {}

    def handle_CAssignment(self, obj: CAssignment):
        if (
            isinstance(obj.lhs, CVariable)
            and isinstance(obj.lhs.variable, SimMemoryVariable)
            and isinstance(obj.lhs.variable.addr, int)
            and isinstance(obj.rhs, CFunctionCall)
            and isinstance(obj.rhs.callee_target, CConstant)
            and obj.rhs.callee_target.value == self.func_addr
        ):
            # found it!
            func_args = obj.rhs.args
            if self.desc.funcname_argidx < len(func_args) and self.desc.libname_argidx < len(func_args):
                funcname_arg = func_args[self.desc.funcname_argidx]
                libname_arg = func_args[self.desc.libname_argidx]
                if isinstance(funcname_arg, CConstant) and isinstance(libname_arg, CConstant):
                    # load two strings
                    funcname, libname = None, None
                    if funcname_arg.type in funcname_arg.reference_values and isinstance(
                        funcname_arg.reference_values[funcname_arg.type].content, bytes
                    ):
                        funcname = funcname_arg.reference_values[funcname_arg.type].content.decode("utf-8")
                    if libname_arg.type in libname_arg.reference_values and isinstance(
                        libname_arg.reference_values[libname_arg.type].content, bytes
                    ):
                        libname = libname_arg.reference_values[libname_arg.type].content.decode("utf-8")

                    if funcname and libname:
                        if obj.lhs.variable.addr in self.assignments:
                            if self.assignments[obj.lhs.variable.addr] != (libname, funcname):
                                _l.warning(
                                    "Observed more than one assignment for variable at %#x.", obj.lhs.variable.addr
                                )
                        else:
                            self.assignments[obj.lhs.variable.addr] = libname, funcname

        return super().handle_CAssignment(obj)


class APIObfuscationFinder(Analysis):
    """
    An analysis that automatically finds API "obfuscation" routines.

    Currently, we support the following API "obfuscation" styles:

    - Type 1: sub_A("dll_name", "api_name") where sub_A ends up calling LoadLibrary.
    - Type 2: GetProcAddress(_, "api_name").
    """

    def __init__(self, variable_kb: KnowledgeBase | None = None):
        self.type1_candidates = []
        self.variable_kb = variable_kb or self.project.kb

        self.analyze()

    def analyze(self):
        self.type1_candidates = self._find_type1()

        if self.type1_candidates:
            for desc in self.type1_candidates:
                type1_deobfuscated = self._analyze_type1(desc.func_addr, desc)
                self.kb.obfuscations.type1_deobfuscated_apis.update(type1_deobfuscated)

        APIObfuscationType2Finder(self.project, self.variable_kb).analyze()
        self.project.analyses[HashLookupAPIDeobfuscator].prep(fail_fast=self._fail_fast)(
            self._hash_lookup_api_deobfuscator_lifter
        )

    def _hash_lookup_api_deobfuscator_lifter(self, func):
        d = self.project.analyses.Decompiler(func)
        return d.clinic

    def _find_type1(self):
        cfg = self.kb.cfgs.get_most_accurate()
        load_library_funcs = []

        if "LoadLibraryA" in self.kb.functions:
            load_library_funcs += list(self.kb.functions.get_by_name("LoadLibraryA"))
        if "LoadLibraryW" in self.kb.functions:
            load_library_funcs += list(self.kb.functions.get_by_name("LoadLibraryW"))
        if "LoadLibrary" in self.kb.functions:
            load_library_funcs += list(self.kb.functions.get_by_name("LoadLibrary"))

        load_library_funcs = [func for func in load_library_funcs if func.is_simprocedure]

        if not load_library_funcs:
            return None

        # find callers of each load library func, up to three callers back
        callgraph = self.kb.functions.callgraph
        candidates = []
        for load_library_func in load_library_funcs:
            subtree = self._build_caller_subtree(callgraph, load_library_func.addr, 3)
            for _, succs in networkx.bfs_successors(subtree, load_library_func.addr):
                for succ_addr in succs:
                    func = self.kb.functions.get_by_addr(succ_addr)
                    likely, info = self._is_likely_type1_func(func, cfg)
                    if likely:
                        candidates.append((func.addr, info))

        descs = []
        for func_addr, info in candidates:
            desc = APIDeobFuncDescriptor(
                APIObfuscationType.TYPE_1,
                func_addr=func_addr,
                libname_argidx=info["libname_arg_idx"],
                funcname_argidx=info["funcname_arg_idx"],
            )
            descs.append(desc)

        return descs

    def _is_likely_type1_func(self, func, cfg):
        if func.prototype is None:
            return False, None
        if len(func.prototype.args) < 2:
            return False, None

        arch = self.project.arch
        valid_apiname_charset = {ord(ch) for ch in (string.ascii_letters + string.digits + "._")}

        # decompile the function to get a prototype with types
        _ = self.project.analyses.Decompiler(func, cfg=cfg)

        char_ptr_args = [
            idx
            for (idx, arg) in enumerate(func.prototype.args)
            if isinstance(arg, SimTypePointer) and isinstance(arg.pts_to, SimTypeChar)
        ]
        if len(char_ptr_args) != 2:
            return False, None

        libname_arg_idx = None
        funcname_arg_idx = None
        # who's calling it?
        caller_addrs = sorted(set(self.kb.functions.callgraph.predecessors(func.addr)))
        for caller_addr in caller_addrs:
            # what arguments are used to call this function with?
            callsite_nodes = [
                pred
                for pred in cfg.get_predecessors(cfg.get_any_node(func.addr))
                if pred.function_address == caller_addr and pred.instruction_addrs
            ]
            observation_points = []
            for callsite_node in callsite_nodes:
                observation_points.append(("insn", callsite_node.instruction_addrs[-1], ObservationPointType.OP_BEFORE))
            rda = self.project.analyses.ReachingDefinitions(
                self.kb.functions[caller_addr],
                observe_all=False,
                observation_points=observation_points,
            )
            for callsite_node in callsite_nodes:
                observ = rda.model.get_observation_by_insn(
                    callsite_node.instruction_addrs[-1],
                    ObservationPointType.OP_BEFORE,
                )
                if observ is None:
                    continue
                args: list[tuple[int, Any]] = []
                for arg_idx, func_arg in enumerate(func.arguments):
                    # FIXME: We are ignoring all non-register function arguments until we see a test case where
                    # FIXME: stack-passing arguments are used
                    if isinstance(func_arg, SimRegArg):
                        reg_offset, reg_size = arch.registers[func_arg.reg_name]
                        try:
                            mv = observ.registers.load(reg_offset, size=reg_size)
                        except SimMemoryMissingError:
                            args.append((arg_idx, claripy.BVV(0xDEADBEEF, self.project.arch.bits)))
                            continue
                        arg_value = mv.one_value()
                        if arg_value is None:
                            arg_value = claripy.BVV(0xDEADBEEF, self.project.arch.bits)
                        args.append((arg_idx, arg_value))

                # the args must have at least one concrete address that points to an initialized memory location
                acceptable_args = True
                arg_strs: list[tuple[int, str]] = []
                for idx, arg in args:
                    if arg is not None and arg.concrete:
                        v = arg.concrete_value
                        section = self.project.loader.find_section_containing(v)
                        if section is not None:
                            # what string is it?
                            max_size = min(64, section.max_addr - v)
                            try:
                                value = self.project.loader.memory.load(v, max_size)
                            except KeyError:
                                acceptable_args = False
                                break
                            if b"\x00" in value:
                                value = value[: value.index(b"\x00")]
                            if not all(ch in valid_apiname_charset for ch in value):
                                acceptable_args = False
                                break
                            arg_strs.append((idx, value.decode("utf-8")))
                if acceptable_args and len(arg_strs) == 2:
                    libname_arg_idx, funcname_arg_idx = None, None
                    for arg_idx, name in arg_strs:
                        if self.is_libname(name):
                            libname_arg_idx = arg_idx
                        elif self.is_apiname(name):
                            funcname_arg_idx = arg_idx

                    if libname_arg_idx is not None and funcname_arg_idx is not None:
                        break

            if libname_arg_idx is not None and funcname_arg_idx is not None:
                break

        if libname_arg_idx is not None and funcname_arg_idx is not None:
            return True, {"libname_arg_idx": libname_arg_idx, "funcname_arg_idx": funcname_arg_idx}
        return False, None

    def _analyze_type1(self, func_addr, desc: APIDeobFuncDescriptor) -> dict[int, tuple[str, str]]:
        cfg = self.kb.cfgs.get_most_accurate()

        assignments: dict[int, tuple[str, str]] = {}

        # get all call sites
        caller_addrs = sorted(set(self.kb.functions.callgraph.predecessors(func_addr)))
        for caller_addr in caller_addrs:
            # decompile the function and get all assignments of the return value of the func at func_addr
            try:
                dec = self.project.analyses.Decompiler(self.kb.functions.get_by_addr(caller_addr), cfg=cfg)
            except Exception:  # pylint:disable=broad-exception-caught
                continue
            if dec.codegen is None:
                continue

            finder = Type1AssignmentFinder(func_addr, desc)
            finder.handle(dec.codegen.cfunc)

            duplicate_addrs = set(assignments.keys()).intersection(set(finder.assignments.keys()))
            if duplicate_addrs:
                # duplicate entries
                _l.warning(
                    "Observed duplicate assignments at the following addresses: %s.",
                    str(map(hex, sorted(duplicate_addrs))),  # pylint:disable=bad-builtin
                )

            assignments.update(finder.assignments)

        return assignments

    @staticmethod
    def _build_caller_subtree(callgraph: networkx.DiGraph, func_addr: int, max_level: int) -> networkx.DiGraph:
        tree = networkx.DiGraph()

        if func_addr not in callgraph:
            return tree

        queue = [(0, func_addr)]
        traversed = {func_addr}
        while queue:
            level, addr = queue.pop(0)
            for pred in callgraph.predecessors(addr):
                if pred not in traversed and level + 1 <= max_level:
                    traversed.add(pred)
                    queue.append((level + 1, pred))
                    tree.add_edge(addr, pred)

        return tree

    @staticmethod
    def is_libname(name: str) -> bool:
        name = name.lower()
        if name in SIM_LIBRARIES:
            return True
        if "." not in name:
            return name + ".dll" in SIM_LIBRARIES or name + ".exe" in SIM_LIBRARIES
        return False

    @staticmethod
    def is_apiname(name: str) -> bool:
        return any(
            not isinstance(lib, SimSyscallLibrary) and lib.has_prototype(name)
            for libs in SIM_LIBRARIES.values()
            for lib in libs
        )


AnalysesHub.register_default("APIObfuscationFinder", APIObfuscationFinder)



# what a silly fulcrum. the SimState engines and the the light ail engines have diverged enough
# that they can no longer be mixed together.
# # simply compose them by encapsulation instead of inheritance.
from __future__ import annotations

__all__ = [
    "AILCallStack",
    "AILMixin",
    "SimEngineAILSimState",
    "ail_call_state",
]

from .engine_successors import AILMixin
from .engine_light import SimEngineAILSimState
from .setup import ail_call_state
from .callstack import AILCallStack



from __future__ import annotations

from angr.ailment.block import Block
from angr.engines.ail.engine_light import SimEngineAILSimState
from angr.engines.successors import SimSuccessors, SuccessorsEngine


class AILMixin(SuccessorsEngine):
    """
    The SuccessorsEngine mixin for providing AIL symbolic execution support
    """

    def process_successors(self, successors: SimSuccessors, block: Block | None = None, **kwargs):
        if not isinstance(successors.addr, tuple):
            super().process_successors(successors, block=block, **kwargs)
            return

        successors.sort = "AIL"
        successors.description = "AIL block"

        subengine = SimEngineAILSimState(self.project, successors)
        subengine.process(self.state, block=block)  # type: ignore

        successors.processed = True



from __future__ import annotations
import claripy

from angr.procedures.java import JavaSimProcedure


class NextInt(JavaSimProcedure):
    __provides__ = (("java.util.Random", "nextInt(int)"),)

    def run(self, obj, bound):  # pylint: disable=arguments-differ,unused-argument
        rand = claripy.BVS("rand", 32)
        self.state.solver.add(rand.UGE(0))
        self.state.solver.add(rand.ULT(bound))
        return rand



from __future__ import annotations
import logging

from angr import ailment
from angr.utils.types import dereference_simtype_by_lib
from angr.sim_type import SimTypeBottom
from angr.calling_conventions import SimRegArg
from .ailgraph_walker import AILGraphWalker

l = logging.getLogger(__name__)


class ReturnMaker(AILGraphWalker):
    """
    Traverse the AILBlock graph of a function and update .ret_exprs of all return statements.
    """

    def __init__(self, ail_manager, arch, function, ail_graph):
        super().__init__(ail_graph, self._handler, replace_nodes=True)
        self.ail_manager = ail_manager
        self.arch = arch
        self.function = function

        self.walk()

    def _next_atom(self) -> int:
        return self.ail_manager.next_atom()

    def _handle_Return(
        self, stmt_idx: int, stmt: ailment.Stmt.Return, block: ailment.Block | None
    ):  # pylint:disable=unused-argument
        if (
            block is not None
            and not stmt.ret_exprs
            and self.function.prototype is not None
            and self.function.prototype.returnty is not None
            and type(self.function.prototype.returnty) is not SimTypeBottom
        ):
            new_stmt = stmt.copy()
            returnty = (
                dereference_simtype_by_lib(self.function.prototype.returnty, self.function.prototype_libname)
                if self.function.prototype_libname
                else self.function.prototype.returnty
            )
            ret_val = self.function.calling_convention.return_val(returnty)
            if isinstance(ret_val, SimRegArg):
                reg = self.arch.registers[ret_val.reg_name]
                new_stmt.ret_exprs.append(
                    ailment.Expr.Register(
                        self._next_atom(),
                        None,
                        reg[0],
                        ret_val.size * self.arch.byte_width,
                        reg_name=self.arch.translate_register_name(reg[0], ret_val.size),
                        ins_addr=stmt.tags["ins_addr"],
                    )
                )
            else:
                l.warning("Unsupported type of return expression %s.", type(ret_val))
            return new_stmt
        return stmt

    def _handler(self, block):
        # we don't need to handle any statement besides Returns
        walker = ailment.AILBlockRewriter(
            update_block=False, expr_handlers={}, stmt_handlers={ailment.statement.Return: self._handle_Return}
        )

        result = walker.walk(block)
        if result is block:
            return None
        return result



from __future__ import annotations

from angr.ailment.statement import DirtyStatement, Statement
from angr.ailment.expression import DirtyExpression, Expression


class DirtyRewriterBase:
    """
    The base class for DirtyStatement and DirtyExpression rewriters.
    """

    __slots__ = (
        "arch",
        "result",
    )

    def __init__(self, dirty: DirtyExpression | DirtyStatement, arch):
        self.arch = arch
        self.result: Expression | Statement | None = (
            self._rewrite_expr(dirty) if isinstance(dirty, DirtyExpression) else self._rewrite_stmt(dirty)
        )

    def _rewrite_stmt(self, dirty: DirtyStatement) -> Statement | None:
        raise NotImplementedError

    def _rewrite_expr(self, dirty: DirtyExpression) -> Expression | None:
        raise NotImplementedError



from __future__ import annotations
import angr

import logging

l = logging.getLogger(name=__name__)


class fflush(angr.SimProcedure):
    # pylint:disable=arguments-differ,unused-argument

    def run(self, fd):
        return 0


fflush_unlocked = fflush



from __future__ import annotations

import logging

import claripy

from angr import sim_options
from angr.state_plugins.sim_event import resource_event
from .base import ExplorationTechnique
from .common import condition_to_lambda

l = logging.getLogger(name=__name__)


class Explorer(ExplorationTechnique):
    """
    Search for up to "num_find" paths that satisfy condition "find", avoiding condition "avoid". Stashes found paths
    into "find_stash' and avoided paths into "avoid_stash".

    The "find" and "avoid" parameters may be any of:

    - An address to find
    - A set or list of addresses to find
    - A function that takes a path and returns whether or not it matches.

    If an angr CFG is passed in as the "cfg" parameter and "find" is either a number or a list or a set, then
    any paths which cannot possibly reach a success state without going through a failure state will be
    preemptively avoided.

    If either the "find" or "avoid" parameter is a function returning a boolean, and a path triggers both conditions,
    it will be added to the find stash, unless "avoid_priority" is set to True.
    """

    def __init__(
        self, find=None, avoid=None, find_stash="found", avoid_stash="avoid", cfg=None, num_find=1, avoid_priority=False
    ):
        super().__init__()
        self.find, static_find = condition_to_lambda(find)
        self.avoid, static_avoid = condition_to_lambda(avoid)
        self.find_stash = find_stash
        self.avoid_stash = avoid_stash
        self.cfg = cfg
        self.ok_blocks = set()
        self.num_find = num_find
        self.avoid_priority = avoid_priority

        # even if avoid or find addresses are not statically known, stop on those that we do know
        self._extra_stop_points = (static_find or set()) | (static_avoid or set())
        self._unknown_stop_points = static_find is None or static_avoid is None
        self._warned_unicorn = False

        # TODO: This is a hack for while CFGFast doesn't handle procedure continuations
        from angr import analyses  # pylint: disable=import-outside-toplevel

        if isinstance(cfg, analyses.CFGFast):
            l.error("CFGFast is currently inappropriate for use with Explorer.")
            l.error("Usage of the CFG has been disabled for this explorer.")
            self.cfg = None

        if self.cfg is not None:
            avoid = static_avoid or set()

            # we need the find addresses to be determined statically
            if not static_find:
                l.error("You must provide at least one numeric 'find' address if you provide a CFG.")
                l.error("Usage of the CFG has been disabled for this explorer.")
                self.cfg = None
                return

            for a in avoid:
                if cfg.model.get_any_node(a) is None:
                    l.warning("'Avoid' address %#x not present in CFG...", a)

            # not a queue but a stack... it's just a worklist!
            queue = []
            for f in static_find:
                nodes = cfg.model.get_all_nodes(f)
                if len(nodes) == 0:
                    l.warning("'Find' address %#x not present in CFG...", f)
                else:
                    queue.extend(nodes)

            seen_nodes = set()
            while len(queue) > 0:
                n = queue.pop()
                if id(n) in seen_nodes:
                    continue
                if n.addr in avoid:
                    continue
                self.ok_blocks.add(n.addr)
                seen_nodes.add(id(n))
                queue.extend(n.predecessors)

            if len(self.ok_blocks) == 0:
                l.error("No addresses could be validated by the provided CFG!")
                l.error("Usage of the CFG has been disabled for this explorer.")
                self.cfg = None
                return

            l.warning("Please be sure that the CFG you have passed in is complete.")
            l.warning("Providing an incomplete CFG can cause viable paths to be discarded!")

    def setup(self, simgr):
        if self.find_stash not in simgr.stashes:
            simgr.stashes[self.find_stash] = []
        if self.avoid_stash not in simgr.stashes:
            simgr.stashes[self.avoid_stash] = []

    def step(self, simgr, stash="active", **kwargs):
        base_extra_stop_points = set(kwargs.pop("extra_stop_points", []))
        return simgr.step(stash=stash, extra_stop_points=base_extra_stop_points | self._extra_stop_points, **kwargs)

    # make it more natural to deal with the intended dataflow
    def filter(self, simgr, state, **kwargs):
        stash = self._filter_inner(state)
        if stash is None:
            return simgr.filter(state, **kwargs)
        return stash

    def _filter_inner(self, state):
        if self._unknown_stop_points and sim_options.UNICORN in state.options and not self._warned_unicorn:
            l.warning("Using unicorn with find/avoid conditions that are a lambda (not a number, set, tuple or list)")
            l.warning("Unicorn may step over states that match the condition (find or avoid) without stopping.")
            self._warned_unicorn = True

        try:
            if self.avoid_priority:
                avoidable = self.avoid(state)
                if avoidable and (avoidable is True or state.addr in avoidable):
                    return self.avoid_stash
            findable = self.find(state)
            if findable and (findable is True or state.addr in findable):
                return self.find_stash
            if not self.avoid_priority:
                avoidable = self.avoid(state)
                if avoidable and (avoidable is True or state.addr in avoidable):
                    return self.avoid_stash
        except claripy.errors.ClaripySolverInterruptError as e:
            resource_event(state, e)
            return "interrupted"

        if (
            self.cfg is not None
            and self.cfg.model.get_any_node(state.addr) is not None
            and state.addr not in self.ok_blocks
        ):
            return self.avoid_stash

        return None

    def complete(self, simgr):
        return len(simgr.stashes[self.find_stash]) >= self.num_find



from __future__ import annotations

import argparse
import logging
import re
from typing import TYPE_CHECKING
from collections.abc import Generator

from rich.syntax import Syntax
from rich.console import Console

import angr
from angr.analyses.decompiler import DECOMPILATION_PRESETS
from angr.analyses.decompiler.structuring import STRUCTURER_CLASSES, DEFAULT_STRUCTURER
from angr.analyses.decompiler.utils import decompile_functions
from angr.utils.formatting import ansi_color_enabled

if TYPE_CHECKING:
    from angr.knowledge_plugins.functions import Function


log = logging.getLogger(__name__)


NUMERIC_ARG_RE = re.compile(r"^(0x)?[a-fA-F0-9]+$")


def parse_function_args(proj: angr.Project, func_args: list[str] | None) -> Generator[Function]:
    """
    Generate a sequence of functions in the project kb by their identifier in func_args.

    :param proj:      Project to query.
    :param func_args: Sequence of function identifiers to query. None for all functions.
    """
    if func_args is None:
        yield from sorted(proj.kb.functions.values(), key=lambda f: f.addr)
        return

    for func_arg in func_args:
        if func_arg in proj.kb.functions:
            yield proj.kb.functions[func_arg]
            continue

        if NUMERIC_ARG_RE.match(func_arg):
            func_addr = int(func_arg, 0)
            if func_addr in proj.kb.functions:
                yield proj.kb.functions[func_addr]
                continue

        log.error('Function "%s" not found', func_arg)


def disassemble(args):
    """
    Disassemble functions.
    """
    loader_main_opts_kwargs = {}
    if args.base_addr is not None:
        loader_main_opts_kwargs["base_addr"] = args.base_addr

    proj = angr.Project(args.binary, auto_load_libs=False, main_opts=loader_main_opts_kwargs)
    proj.analyses.CFG(normalize=True, data_references=True)

    for func in parse_function_args(proj, args.functions):
        try:
            if func.is_plt or func.is_syscall or func.is_alignment or func.is_simprocedure:
                continue
            func.pp(show_bytes=True, min_edge_depth=10)
        except Exception as e:  # pylint:disable=broad-exception-caught
            if not args.catch_exceptions:
                raise
            log.exception(e)


def decompile(args):
    """
    Decompile functions.
    """
    decompilation = decompile_functions(
        args.binary,
        functions=args.functions,
        structurer=args.structurer,
        catch_errors=args.catch_exceptions,
        show_casts=not args.no_casts,
        base_address=args.base_addr,
        preset=args.preset,
    )

    # Determine if we should use syntax highlighting
    should_highlight = ansi_color_enabled and not args.no_colors

    if should_highlight:
        try:
            console = Console()
            syntax = Syntax(decompilation, "c", theme=args.theme, line_numbers=False)
            console.print(syntax)
        # pylint: disable=broad-exception-caught
        except Exception as e:
            log.warning("Syntax highlighting failed: %s", e)
            # Fall back to plain text if syntax highlighting fails
            print(decompilation)
    else:
        print(decompilation)


def main():
    parser = argparse.ArgumentParser(description="The angr CLI allows you to decompile and analyze binaries.")
    parser.add_argument("--version", action="version", version=angr.__version__)
    parser.add_argument("binary", help="The path to the binary to analyze.")
    parser.add_argument(
        "--catch-exceptions",
        help="""
        Catch exceptions during analysis. The scope of error handling may depend on the command used for analysis.
        If multiple functions are specified for analysis, each function will be handled individually.""",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--base-addr",
        help="""
        The base address of the binary. This is useful when the binary is loaded at a different address than the one
        specified in the ELF header.""",
        type=lambda x: int(x, 0),
        default=None,
    )
    subparsers = parser.add_subparsers(metavar="command", required=True)

    decompile_cmd_parser = subparsers.add_parser("decompile", aliases=["dec"], help=decompile.__doc__)
    decompile_cmd_parser.set_defaults(func=decompile)
    decompile_cmd_parser.add_argument(
        "--structurer",
        help="The structuring algorithm to use for decompilation.",
        choices=STRUCTURER_CLASSES.keys(),
        default=DEFAULT_STRUCTURER.NAME,
    )
    decompile_cmd_parser.add_argument(
        "--no-casts",
        help="Do not show type casts in the decompiled output.",
        action="store_true",
        default=False,
    )
    decompile_cmd_parser.add_argument(
        "--preset",
        help="The configuration preset to use for decompilation.",
        choices=DECOMPILATION_PRESETS,
        default="default",
    )
    decompile_cmd_parser.add_argument(
        "--functions",
        help="""
        The functions to decompile. Functions can either be expressed as names found in the
        symbols of the binary or as addresses like: 0x401000.""",
        nargs="+",
    )
    decompile_cmd_parser.add_argument(
        "--no-colors",
        help="Disable syntax highlighting in the decompiled output.",
        action="store_true",
        default=False,
    )
    decompile_cmd_parser.add_argument(
        "--theme",
        help="The syntax highlighting theme to use (only if rich is installed and colors are enabled).",
        default="dracula",
    )

    disassemble_cmd_parser = subparsers.add_parser("disassemble", aliases=["dis"], help=disassemble.__doc__)
    disassemble_cmd_parser.set_defaults(func=disassemble)
    disassemble_cmd_parser.add_argument(
        "--functions",
        help="""
        The functions to disassemble. Functions can either be expressed as names found in the
        symbols of the binary or as addresses like: 0x401000.""",
        nargs="+",
    )

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



# pylint:disable=missing-class-docstring,no-self-use
from __future__ import annotations
from angr.ailment.expression import BinaryOp, Const

from .base import PeepholeOptimizationExprBase


class ShlToMul(PeepholeOptimizationExprBase):
    __slots__ = ()

    NAME = "a << A => a * (2 ** A)"
    expr_classes = (BinaryOp,)  # all expressions are allowed

    def optimize(self, expr: BinaryOp, **kwargs):
        if expr.op == "Shl" and isinstance(expr.operands[1], Const):
            mul_amount = Const(None, None, 2 ** expr.operands[1].value, expr.operands[0].bits)
            return BinaryOp(
                expr.idx,
                "Mul",
                [expr.operands[0], mul_amount],
                expr.signed,
                **expr.tags,
            )

        return None



# pylint:disable=arguments-renamed,too-many-boolean-expressions,no-self-use,unused-argument
from __future__ import annotations
from typing import Any
from collections.abc import Callable
from collections import defaultdict

from archinfo import Endness
from angr.ailment.expression import (
    Const,
    Register,
    Expression,
    StackBaseOffset,
    Convert,
    BinaryOp,
    VirtualVariable,
    Phi,
    UnaryOp,
    VirtualVariableCategory,
)
from angr.ailment.statement import ConditionalJump, Jump, Assignment
import claripy

from angr.utils.bits import zeroextend_on_demand
from angr.engines.light import SimEngineNostmtAIL
from angr.storage.memory_mixins import (
    SimpleInterfaceMixin,
    DefaultFillerMixin,
    PagedMemoryMixin,
    UltraPagesMixin,
)
from angr.code_location import CodeLocation
from angr.errors import SimMemoryMissingError
from .optimization_pass import OptimizationPass, OptimizationPassStage


def _make_binop(compute: Callable[[claripy.ast.BV, claripy.ast.BV], claripy.ast.BV]):
    def inner(self, expr: BinaryOp) -> claripy.ast.BV | None:
        a, b = self._expr(expr.operands[0]), self._expr(expr.operands[1])
        if a is None or b is None:
            return None
        try:
            return compute(a, b)
        except (ZeroDivisionError, claripy.ClaripyZeroDivisionError):
            return None

    return inner


class FasterMemory(
    SimpleInterfaceMixin,
    DefaultFillerMixin,
    UltraPagesMixin,
    PagedMemoryMixin,
):
    """
    A fast memory model used in InlinedStringTransformationState.
    """


class InlinedStringTransformationState:
    """
    The abstract state used in InlinedStringTransformationAILEngine.
    """

    def __init__(self, project):
        self.arch = project.arch
        self.project = project

        self.registers = FasterMemory(memory_id="reg")
        self.memory = FasterMemory(memory_id="mem")
        self.virtual_variables = {}

        self.registers.set_state(self)
        self.memory.set_state(self)

    def _get_weakref(self):
        return self

    def reg_store(self, reg: Register, value: claripy.ast.BV) -> None:
        self.registers.store(
            reg.reg_offset, value, size=value.size() // self.arch.byte_width, endness=str(self.arch.register_endness)
        )

    def reg_load(self, reg: Register) -> claripy.ast.BV | None:
        try:
            return self.registers.load(
                reg.reg_offset, size=reg.size, endness=self.arch.register_endness, fill_missing=False
            )
        except SimMemoryMissingError:
            return None

    def mem_store(self, addr: int, value: claripy.ast.Bits, endness: str) -> None:
        self.memory.store(addr, value, size=value.size() // self.arch.byte_width, endness=endness)

    def mem_load(self, addr: int, size: int, endness) -> claripy.ast.BV | None:
        try:
            return self.memory.load(addr, size=size, endness=str(endness), fill_missing=False)
        except SimMemoryMissingError:
            return None

    def vvar_store(self, vvar: VirtualVariable, value: claripy.ast.Bits | None) -> None:
        self.virtual_variables[vvar.varid] = value

    def vvar_load(self, vvar: VirtualVariable) -> claripy.ast.BV | None:
        if vvar.varid in self.virtual_variables:
            return self.virtual_variables[vvar.varid]
        return None


class InlinedStringTransformationAILEngine(
    SimEngineNostmtAIL[InlinedStringTransformationState, claripy.ast.BV | None, None, None],
):
    """
    A simple AIL execution engine
    """

    def __init__(self, project, nodes: dict[int, Any], start: int, end: int, step_limit: int):
        super().__init__(project)
        self.nodes: dict[int, Any] = nodes
        self.start: int = start
        self.end: int = end
        self.step_limit: int = step_limit

        self.STACK_BASE = 0x7FFF_FFF0 if self.arch.bits == 32 else 0x7FFF_FFFF_F000
        self.MASK = 0xFFFF_FFFF if self.arch.bits == 32 else 0xFFFF_FFFF_FFFF_FFFF

        state = InlinedStringTransformationState(project)
        self.stack_accesses: defaultdict[int, list[tuple[str, CodeLocation, claripy.ast.Bits]]] = defaultdict(list)
        self.finished: bool = False

        i = 0
        self.last_pc = None
        self.pc = self.start
        while i < self.step_limit:
            if self.pc not in self.nodes:
                # jumped to a node that we do not know about
                break
            block = self.nodes[self.pc]
            self.process(state, block=block, whitelist=None)
            if self.pc is None:
                # not sure where to jump...
                break
            if self.pc == self.end:
                # we reach the end of execution!
                self.finished = True
                break
            i += 1

    def _top(self, bits):
        assert False, "Should not be reachable"

    def _is_top(self, expr):
        assert False, "Should not be reachable"

    def _process_block_end(self, block, stmt_data, whitelist):
        pass

    def _process_address(self, addr: Expression) -> tuple[int, str] | None:
        if isinstance(addr, Const):
            assert isinstance(addr.value, int)
            return addr.value, "mem"
        if isinstance(addr, StackBaseOffset):
            return (addr.offset + self.STACK_BASE) & self.MASK, "stack"
        if (
            isinstance(addr, UnaryOp)
            and addr.op == "Reference"
            and isinstance(addr.operand, VirtualVariable)
            and addr.operand.was_stack
        ):
            return (addr.operand.stack_offset + self.STACK_BASE) & self.MASK, "stack"
        if (
            isinstance(addr, BinaryOp)
            and addr.op in {"Add", "Sub"}
            and isinstance(addr.operands[0], (StackBaseOffset, UnaryOp, Const))
        ):
            v0_and_type = self._process_address(addr.operands[0])
            if v0_and_type is not None:
                v0 = v0_and_type[0]
                v1 = self._expr(addr.operands[1])
                if isinstance(v1, claripy.ast.Bits) and v1.concrete:
                    if addr.op == "Add":
                        return (v0 + v1.concrete_value) & self.MASK, "stack"
                    if addr.op == "Sub":
                        return (v0 - v1.concrete_value) & self.MASK, "stack"
                    raise NotImplementedError("Unreachable")
        return None

    def _handle_stmt_Assignment(self, stmt):
        if isinstance(stmt.dst, VirtualVariable):
            if stmt.dst.was_reg:
                val = self._expr(stmt.src)
                if isinstance(val, claripy.ast.Bits):
                    self.state.vvar_store(stmt.dst, val)
            elif stmt.dst.was_stack:
                addr = (stmt.dst.stack_offset + self.STACK_BASE) & self.MASK
                val = self._expr(stmt.src)
                if isinstance(val, claripy.ast.BV):
                    self.state.mem_store(addr, val, self.arch.memory_endness)
                    # log it
                    for i in range(val.size() // self.arch.byte_width):
                        byte_off = i
                        if self.arch.memory_endness == Endness.LE:
                            byte_off = val.size() // self.arch.byte_width - i - 1
                        self.stack_accesses[addr + i].append(("store", self._codeloc(), val.get_byte(byte_off)))

    def _handle_stmt_Store(self, stmt):
        addr_and_type = self._process_address(stmt.addr)
        if addr_and_type is not None:
            addr, addr_type = addr_and_type
            val = self._expr(stmt.data)
            if isinstance(val, claripy.ast.BV):
                self.state.mem_store(addr, val, stmt.endness)
                # log it
                if addr_type == "stack":
                    for i in range(val.size() // self.arch.byte_width):
                        byte_off = i
                        if self.arch.memory_endness == Endness.LE:
                            byte_off = val.size() // self.arch.byte_width - i - 1
                        self.stack_accesses[addr + i].append(("store", self._codeloc(), val.get_byte(byte_off)))

    def _handle_stmt_Jump(self, stmt):
        self.last_pc = self.pc
        if isinstance(stmt.target, Const):
            self.pc = stmt.target.value
        else:
            self.pc = None

    def _handle_stmt_ConditionalJump(self, stmt):
        self.last_pc = self.pc
        self.pc = None
        if isinstance(stmt.true_target, Const) and isinstance(stmt.false_target, Const):
            cond = self._expr(stmt.condition)
            if cond is not None:
                if isinstance(cond, claripy.ast.Bits) and cond.concrete_value == 1:
                    self.pc = stmt.true_target.value
                elif isinstance(cond, claripy.ast.Bits) and cond.concrete_value == 0:
                    self.pc = stmt.false_target.value

    def _handle_expr_Const(self, expr):
        if isinstance(expr.value, int):
            return claripy.BVV(expr.value, expr.bits)
        return None

    def _handle_expr_Load(self, expr):
        addr_and_type = self._process_address(expr.addr)
        if addr_and_type is not None:
            addr, addr_type = addr_and_type
            v = self.state.mem_load(addr, expr.size, expr.endness)
            # log it
            if addr_type == "stack" and isinstance(v, claripy.ast.BV):
                for i in range(expr.size):
                    byte_off = i
                    if self.arch.memory_endness == Endness.LE:
                        byte_off = expr.size - i - 1
                    self.stack_accesses[addr + i].append(("load", self._codeloc(), v.get_byte(byte_off)))
            return v
        return None

    def _handle_expr_Register(self, expr: Register):
        return self.state.reg_load(expr)

    def _handle_expr_VirtualVariable(self, expr: VirtualVariable):
        if expr.was_stack:
            addr = (expr.stack_offset + self.STACK_BASE) & self.MASK
            v = self.state.mem_load(addr, expr.size, self.arch.memory_endness)
            if v is not None:
                # log it
                for i in range(expr.size):
                    byte_off = i
                    if self.arch.memory_endness == Endness.LE:
                        byte_off = expr.size - i - 1
                    self.stack_accesses[addr + i].append(("load", self._codeloc(), v.get_byte(byte_off)))
            return v
        if expr.was_reg:
            return self.state.vvar_load(expr)
        return None

    def _handle_expr_Phi(self, expr: Phi):
        for src, vvar in expr.src_and_vvars:
            if src[0] == self.last_pc and vvar is not None:
                return self.state.vvar_load(vvar)
        return None

    def _handle_unop_Neg(self, expr: UnaryOp):
        v = self._expr(expr.operand)
        if isinstance(v, claripy.ast.Bits):
            return -v
        return None

    def _handle_unop_Not(self, expr: UnaryOp):
        v = self._expr(expr.operand)
        if isinstance(v, claripy.ast.Bits):
            return ~v
        return None

    def _handle_unop_BitwiseNeg(self, expr: UnaryOp):
        v = self._expr(expr.operand)
        if isinstance(v, claripy.ast.Bits):
            return ~v
        return None

    def _handle_unop_Default(self, expr: UnaryOp):
        return None

    _handle_unop_Clz = _handle_unop_Default
    _handle_unop_Ctz = _handle_unop_Default
    _handle_unop_Dereference = _handle_unop_Default
    _handle_unop_Reference = _handle_unop_Default
    _handle_unop_GetMSBs = _handle_unop_Default
    _handle_unop_unpack = _handle_unop_Default
    _handle_unop_Sqrt = _handle_unop_Default
    _handle_unop_RSqrtEst = _handle_unop_Default

    def _handle_expr_Convert(self, expr: Convert):
        v = self._expr(expr.operand)
        if isinstance(v, claripy.ast.Bits):
            if expr.to_bits > expr.from_bits:
                if not expr.is_signed:
                    return claripy.ZeroExt(expr.to_bits - expr.from_bits, v)
                return claripy.SignExt(expr.to_bits - expr.from_bits, v)
            if expr.to_bits < expr.from_bits:
                return claripy.Extract(expr.to_bits - 1, 0, v)
            return v
        return None

    def _handle_binop_CmpEQ(self, expr):
        op0, op1 = self._expr(expr.operands[0]), self._expr(expr.operands[1])
        if isinstance(op0, claripy.ast.Bits) and isinstance(op1, claripy.ast.Bits) and op0.concrete and op1.concrete:
            return claripy.BVV(1, 1) if op0.concrete_value == op1.concrete_value else claripy.BVV(0, 1)
        return None

    def _handle_binop_CmpNE(self, expr):
        op0, op1 = self._expr(expr.operands[0]), self._expr(expr.operands[1])
        if isinstance(op0, claripy.ast.Bits) and isinstance(op1, claripy.ast.Bits) and op0.concrete and op1.concrete:
            return claripy.BVV(1, 1) if op0.concrete_value != op1.concrete_value else claripy.BVV(0, 1)
        return None

    def _handle_binop_CmpLT(self, expr):
        op0, op1 = self._expr(expr.operands[0]), self._expr(expr.operands[1])
        if isinstance(op0, claripy.ast.Bits) and isinstance(op1, claripy.ast.Bits) and op0.concrete and op1.concrete:
            return claripy.BVV(1, 1) if op0.concrete_value < op1.concrete_value else claripy.BVV(0, 1)
        return None

    def _handle_binop_CmpLE(self, expr):
        op0, op1 = self._expr(expr.operands[0]), self._expr(expr.operands[1])
        if isinstance(op0, claripy.ast.Bits) and isinstance(op1, claripy.ast.Bits) and op0.concrete and op1.concrete:
            return claripy.BVV(1, 1) if op0.concrete_value <= op1.concrete_value else claripy.BVV(0, 1)
        return None

    def _handle_binop_CmpGT(self, expr):
        op0, op1 = self._expr(expr.operands[0]), self._expr(expr.operands[1])
        if isinstance(op0, claripy.ast.Bits) and isinstance(op1, claripy.ast.Bits) and op0.concrete and op1.concrete:
            return claripy.BVV(1, 1) if op0.concrete_value > op1.concrete_value else claripy.BVV(0, 1)
        return None

    def _handle_binop_CmpGE(self, expr):
        op0, op1 = self._expr(expr.operands[0]), self._expr(expr.operands[1])
        if isinstance(op0, claripy.ast.Bits) and isinstance(op1, claripy.ast.Bits) and op0.concrete and op1.concrete:
            return claripy.BVV(1, 1) if op0.concrete_value >= op1.concrete_value else claripy.BVV(0, 1)
        return None

    def _handle_binop_CmpORD(self, expr):
        return None

    def _handle_stmt_Call(self, stmt):
        pass

    def _handle_expr_Call(self, expr):
        pass

    def _handle_expr_BasePointerOffset(self, expr):
        return None

    def _handle_expr_DirtyExpression(self, expr):
        return None

    def _handle_expr_ITE(self, expr):
        return None

    def _handle_expr_MultiStatementExpression(self, expr):
        return None

    def _handle_expr_Reinterpret(self, expr):
        return None

    def _handle_expr_StackBaseOffset(self, expr):
        return None

    def _handle_expr_Tmp(self, expr):
        try:
            return self.tmps[expr.tmp_idx]
        except KeyError:
            return None

    def _handle_expr_VEXCCallExpression(self, expr):
        return None

    def _handle_binop_Default(self, expr):
        self._expr(expr.operands[0])
        self._expr(expr.operands[1])

    _handle_binop_Add = _make_binop(lambda a, b: a + b)
    _handle_binop_And = _make_binop(lambda a, b: a & b)
    _handle_binop_Concat = _make_binop(lambda a, b: a.concat(b))
    _handle_binop_Div = _make_binop(lambda a, b: a // b)
    _handle_binop_LogicalAnd = _make_binop(lambda a, b: a & b)
    _handle_binop_LogicalOr = _make_binop(lambda a, b: a | b)
    _handle_binop_Mod = _make_binop(lambda a, b: a % b)
    _handle_binop_Mul = _make_binop(lambda a, b: a * b)
    _handle_binop_Or = _make_binop(lambda a, b: a | b)
    _handle_binop_Rol = _make_binop(lambda a, b: claripy.RotateLeft(a, zeroextend_on_demand(a, b)))
    _handle_binop_Ror = _make_binop(lambda a, b: claripy.RotateRight(a, zeroextend_on_demand(a, b)))
    _handle_binop_Sar = _make_binop(lambda a, b: a >> zeroextend_on_demand(a, b))
    _handle_binop_Shl = _make_binop(lambda a, b: a << zeroextend_on_demand(a, b))
    _handle_binop_Shr = _make_binop(lambda a, b: a.LShR(zeroextend_on_demand(a, b)))
    _handle_binop_Sub = _make_binop(lambda a, b: a - b)
    _handle_binop_Xor = _make_binop(lambda a, b: a ^ b)

    def _handle_binop_Mull(self, expr):
        a, b = self._expr(expr.operands[0]), self._expr(expr.operands[1])
        if a is None or b is None:
            return None
        xt = a.size()
        if expr.signed:
            return a.sign_extend(xt) * b.sign_extend(xt)
        return a.zero_extend(xt) * b.zero_extend(xt)

    _handle_binop_AddF = _handle_binop_Default
    _handle_binop_AddV = _handle_binop_Default
    _handle_binop_Carry = _handle_binop_Default
    _handle_binop_CmpF = _handle_binop_Default
    _handle_binop_DivF = _handle_binop_Default
    _handle_binop_DivV = _handle_binop_Default
    _handle_binop_InterleaveLOV = _handle_binop_Default
    _handle_binop_InterleaveHIV = _handle_binop_Default
    _handle_binop_CasCmpEQ = _handle_binop_Default
    _handle_binop_CasCmpNE = _handle_binop_Default
    _handle_binop_ExpCmpNE = _handle_binop_Default
    _handle_binop_SarNV = _handle_binop_Default
    _handle_binop_ShrNV = _handle_binop_Default
    _handle_binop_ShlNV = _handle_binop_Default
    _handle_binop_CmpEQV = _handle_binop_Default
    _handle_binop_CmpNEV = _handle_binop_Default
    _handle_binop_CmpGEV = _handle_binop_Default
    _handle_binop_CmpGTV = _handle_binop_Default
    _handle_binop_CmpLEV = _handle_binop_Default
    _handle_binop_CmpLTV = _handle_binop_Default
    _handle_binop_MulF = _handle_binop_Default
    _handle_binop_MulV = _handle_binop_Default
    _handle_binop_MulHiV = _handle_binop_Default
    _handle_binop_SBorrow = _handle_binop_Default
    _handle_binop_SCarry = _handle_binop_Default
    _handle_binop_SubF = _handle_binop_Default
    _handle_binop_SubV = _handle_binop_Default
    _handle_binop_MinV = _handle_binop_Default
    _handle_binop_MaxV = _handle_binop_Default
    _handle_binop_QAddV = _handle_binop_Default
    _handle_binop_QNarrowBinV = _handle_binop_Default
    _handle_binop_PermV = _handle_binop_Default
    _handle_binop_Set = _handle_binop_Default


class InlineStringTransformationDescriptor:
    """
    Describes an instance of inline string transformation.
    """

    def __init__(self, store_block, loop_body, stack_accesses, beginning_stack_offset):
        self.store_block = store_block
        self.loop_body = loop_body
        self.stack_accesses = stack_accesses
        self.beginning_stack_offset = beginning_stack_offset


class InlinedStringTransformationSimplifier(OptimizationPass):
    """
    Simplifies inlined string transformation routines.
    """

    ARCHES = None
    PLATFORMS = None
    STAGE = OptimizationPassStage.AFTER_GLOBAL_SIMPLIFICATION
    NAME = "Simplify string transformations"
    DESCRIPTION = "Simplify string transformations that are commonly used in obfuscated functions."

    def __init__(self, func, **kwargs):
        super().__init__(func, **kwargs)
        self.analyze()

    def _check(self):
        string_transformation_descs = self._find_string_transformation_loops()

        return bool(string_transformation_descs), {"descs": string_transformation_descs}

    def _analyze(self, cache=None):
        if not cache or "descs" not in cache:
            return

        for desc in cache["descs"]:
            desc: InlineStringTransformationDescriptor

            # remove the original statements
            skip_stmt_indices = set()
            for stack_accesses in desc.stack_accesses:
                # the first element is the initial storing statement
                codeloc = stack_accesses[0][1]
                assert codeloc.block_addr == desc.store_block.addr
                skip_stmt_indices.add(codeloc.stmt_idx)
            new_statements = [
                stmt for idx, stmt in enumerate(desc.store_block.statements) if idx not in skip_stmt_indices
            ]

            # add new statements
            store_statements = []
            for off, stack_accesses in enumerate(desc.stack_accesses):
                # the last element is the final storing statement
                new_value_ast = stack_accesses[-1][2]
                new_value = Const(None, None, new_value_ast.concrete_value, self.project.arch.byte_width)
                stmt = Assignment(
                    None,
                    VirtualVariable(
                        None,
                        self.vvar_id_start,
                        self.project.arch.bits,
                        category=VirtualVariableCategory.STACK,
                        oident=desc.beginning_stack_offset + off,
                        ins_addr=desc.store_block.addr + desc.store_block.original_size - 1,
                    ),
                    new_value,
                    ins_addr=desc.store_block.addr + desc.store_block.original_size - 1,
                )
                self.vvar_id_start += 1
                store_statements.append(stmt)
            if new_statements and isinstance(new_statements[-1], (ConditionalJump, Jump)):
                new_statements = new_statements[:-1] + store_statements + new_statements[-1:]
            else:
                new_statements += store_statements

            new_store_block = desc.store_block.copy(statements=new_statements)
            self._update_block(desc.store_block, new_store_block)

            # remote the loop node
            # since the loop node has exactly one external predecessor and one external successor, we can get rid of it
            assert self.out_graph is not None
            pred = next(iter(nn for nn in self.out_graph.predecessors(desc.loop_body) if nn is not desc.loop_body))
            succ = next(iter(nn for nn in self.out_graph.successors(desc.loop_body) if nn is not desc.loop_body))

            self.out_graph.remove_node(desc.loop_body)
            self.out_graph.add_edge(pred, succ)

            if pred.statements and isinstance(pred.statements[-1], ConditionalJump):
                pred.statements[-1] = Jump(
                    None,
                    Const(None, None, succ.addr, self.project.arch.bits),
                    succ.idx,
                    **pred.statements[-1].tags,
                )

    def _find_string_transformation_loops(self):
        # find self loops
        self_loops = []
        assert self._graph is not None
        for node in self._graph.nodes:
            preds = list(self._graph.predecessors(node))
            succs = list(self._graph.successors(node))
            if len(preds) == 2 and len(succs) == 2 and node in preds and node in succs:
                pred = next(iter(nn for nn in preds if nn is not node))
                succ = next(iter(nn for nn in succs if nn is not node))
                if (self._graph.out_degree[pred] == 1 and self._graph.in_degree[succ] == 1) or (
                    self._graph.out_degree[pred] == 2
                    and self._graph.in_degree[succ] == 2
                    and self._graph.has_edge(pred, succ)
                ):
                    # found it
                    self_loops.append(node)

        if not self_loops:
            return []

        descs = []
        for loop_node in self_loops:
            pred = next(iter(nn for nn in self._graph.predecessors(loop_node) if nn is not loop_node))
            succ = next(iter(nn for nn in self._graph.successors(loop_node) if nn is not loop_node))
            engine = InlinedStringTransformationAILEngine(
                self.project, {pred.addr: pred, loop_node.addr: loop_node}, pred.addr, succ.addr, 1024
            )
            if engine.finished:
                # find the longest slide where the stack accesses are like the following:
                #   "store", code_location_a, value_a
                #   "load", code_location_b, value_a
                #   "store", code_location_b, value_b
                # where value_a and value_b may be the same
                candidate_stack_addrs = []
                for stack_addr in sorted(engine.stack_accesses.keys()):
                    stack_accesses = engine.stack_accesses[stack_addr]
                    if len(stack_accesses) == 3:
                        item0, item1, item2 = stack_accesses
                        if (
                            item0[0] == "store"
                            and item1[0] == "load"
                            and item2[0] == "store"
                            and item0[1] != item1[1]
                            and item1[1] == item2[1]
                            and item0[2] is item1[2]
                        ):
                            # found one!
                            candidate_stack_addrs.append(stack_addr)

                if (
                    len(candidate_stack_addrs) >= 2
                    and candidate_stack_addrs[-1] == candidate_stack_addrs[0] + len(candidate_stack_addrs) - 1
                ):
                    filtered_stack_accesses = [engine.stack_accesses[a] for a in candidate_stack_addrs]
                    stack_offset = candidate_stack_addrs[0] - engine.STACK_BASE
                    info = InlineStringTransformationDescriptor(pred, loop_node, filtered_stack_accesses, stack_offset)
                    descs.append(info)

        return descs



from __future__ import annotations
import claripy

import angr


class tolower(angr.SimProcedure):
    # pylint: disable=arguments-differ
    def run(self, c):
        return claripy.If(claripy.And(c >= 65, c <= 90), c + 32, c)  # A - Z



from __future__ import annotations
import angr


class getpass(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, prompt):
        # write out the prompt
        self.inline_call(angr.SIM_PROCEDURES["libc"]["puts"], prompt)

        # malloc a buffer
        buf = self.inline_call(angr.SIM_PROCEDURES["libc"]["malloc"], 1024).ret_expr

        # read into the buffer
        self.inline_call(angr.SIM_PROCEDURES["posix"]["read"], 0, buf, 1024)

        # return the buffer
        return buf



# pylint:disable=unused-argument
from __future__ import annotations
import logging
from collections import namedtuple
from collections.abc import Generator
from typing import Any, TYPE_CHECKING
from enum import Enum

import networkx

import angr.ailment as ailment

from angr.analyses.decompiler import RegionIdentifier
from angr.analyses.decompiler.ailgraph_walker import AILGraphWalker
from angr.analyses.decompiler.condition_processor import ConditionProcessor
from angr.analyses.decompiler.goto_manager import Goto, GotoManager
from angr.analyses.decompiler.structuring import RecursiveStructurer, SAILRStructurer
from angr.analyses.decompiler.utils import add_labels, remove_edges_in_ailgraph, is_empty_node
from angr.analyses.decompiler.counters import ControlFlowStructureCounter
from angr.project import Project

if TYPE_CHECKING:
    from angr.knowledge_plugins.functions import Function
    from angr.sim_variable import SimVariable
    from angr.analyses.decompiler.stack_item import StackItem


_l = logging.getLogger(__name__)


BlockCache = namedtuple("BlockCache", ("rd", "prop"))


class MultipleBlocksException(Exception):
    """
    An exception that is raised in _get_block() where multiple blocks satisfy the criteria but only one block was
    requested.
    """


class OptimizationPassStage(Enum):
    """
    Enums about optimization pass stages.

    Note that the region identification pass (RegionIdentifier) may modify existing AIL blocks *without updating the
    topology of the original AIL graph*. For example, loop successor refinement may modify create a new AIL block with
    an artificial address, and alter existing jump targets of jump statements and conditional jump statements to point
    to this new block. However, loop successor refinement does not update the topology of the original AIL graph, which
    means this new AIL block does not exist in the original AIL graph. As a result, until this behavior of
    RegionIdentifier changes in the future, DURING_REGION_IDENTIFICATION optimization passes should not modify existing
    jump targets.
    """

    AFTER_AIL_GRAPH_CREATION = 0
    BEFORE_SSA_LEVEL0_TRANSFORMATION = 1
    AFTER_SINGLE_BLOCK_SIMPLIFICATION = 2
    BEFORE_SSA_LEVEL1_TRANSFORMATION = 3
    AFTER_SSA_LEVEL1_TRANSFORMATION = 4
    AFTER_MAKING_CALLSITES = 5
    AFTER_GLOBAL_SIMPLIFICATION = 6
    BEFORE_VARIABLE_RECOVERY = 7
    AFTER_VARIABLE_RECOVERY = 8
    BEFORE_REGION_IDENTIFICATION = 9
    DURING_REGION_IDENTIFICATION = 10
    AFTER_STRUCTURING = 11


class BaseOptimizationPass:
    """
    The base class for any optimization pass.
    """

    ARCHES = []  # strings of supported architectures
    PLATFORMS = []  # strings of supported platforms. Can be one of the following: "win32", "linux"
    STAGE: OptimizationPassStage  # Specifies when this optimization pass should be executed
    STRUCTURING: list[str] | None = (
        None  # specifies if this optimization pass is specific to a certain structuring algorithm
    )
    NAME = "N/A"
    DESCRIPTION = "N/A"

    def __init__(self, func):
        self._func: Function = func

    @property
    def project(self) -> Project:
        assert self._func.project is not None
        return self._func.project

    @property
    def kb(self):
        return self.project.kb

    def analyze(self):
        ret, cache = self._check()
        if ret:
            self._analyze(cache=cache)

    def _check(self):
        """
        Check if this optimization applies to this function.

        :returns: a tuple of (does_apply, cache) where cache is a way to pass
                  information to _analyze so it does not have to be recalculated
        """
        raise NotImplementedError

    def _analyze(self, cache=None):
        """
        Run the analysis.

        :param cache: information passed from _check so it does not have to be
                      recalculated
        :returns: None
        """
        raise NotImplementedError


class OptimizationPass(BaseOptimizationPass):
    """
    The base class for any function-level graph optimization pass.
    """

    _graph: networkx.DiGraph

    def __init__(
        self,
        func,
        *,
        graph,
        blocks_by_addr=None,
        blocks_by_addr_and_idx=None,
        variable_kb=None,
        region_identifier=None,
        reaching_definitions=None,
        vvar_id_start: int = 0,
        entry_node_addr=None,
        scratch: dict[str, Any] | None = None,
        force_loop_single_exit: bool = True,
        refine_loops_with_single_successor: bool = False,
        complete_successors: bool = False,
        avoid_vvar_ids: set[int] | None = None,
        arg_vvars: dict[int, tuple[ailment.Expr.VirtualVariable, SimVariable]] | None = None,
        peephole_optimizations=None,
        stack_pointer_tracker=None,
        notes: dict | None = None,
        **kwargs,
    ):
        super().__init__(func)
        # self._blocks is just a cache
        self._blocks_by_addr: dict[int, set[ailment.Block]] = blocks_by_addr or {}
        self._blocks_by_addr_and_idx: dict[tuple[int, int | None], ailment.Block] = blocks_by_addr_and_idx or {}
        self._graph = graph
        self._variable_kb = variable_kb
        self._ri = region_identifier
        self._rd = reaching_definitions
        self._scratch = scratch if scratch is not None else {}
        self._new_block_addrs = set()
        self._arg_vvars = arg_vvars
        self.vvar_id_start = vvar_id_start
        self.entry_node_addr: tuple[int, int | None] = (
            entry_node_addr if entry_node_addr is not None else (func.addr, None)
        )
        self._force_loop_single_exit = force_loop_single_exit
        self._refine_loops_with_single_successor = refine_loops_with_single_successor
        self._complete_successors = complete_successors
        self._avoid_vvar_ids = avoid_vvar_ids or set()
        self._peephole_optimizations = peephole_optimizations
        self._stack_pointer_tracker = stack_pointer_tracker
        self.notes = notes if notes is not None else {}

        # output
        self.out_graph: networkx.DiGraph | None = None
        self.stack_items: dict[int, StackItem] = {}

    @property
    def blocks_by_addr(self) -> dict[int, set[ailment.Block]]:
        return self._blocks_by_addr

    @property
    def blocks_by_addr_and_idx(self) -> dict[tuple[int, int | None], ailment.Block]:
        return self._blocks_by_addr_and_idx

    #
    # Util methods
    #

    def bfs_nodes(self, depth: int | None = None, start_node: ailment.Block | None = None) -> Generator[ailment.Block]:
        seen = set()

        if start_node is None:
            start_node = self._get_block(self._func.addr)
        if start_node is None:
            return

        queue = [(0, start_node)]
        while queue:
            node_depth, node = queue.pop(0)
            if node in seen:
                continue
            seen.add(node)

            yield node

            if depth is not None and node_depth >= depth:
                continue

            for succ in sorted(self._graph.successors(node), key=lambda x: (x.addr, x.idx if hasattr(x, "idx") else 0)):
                if succ not in seen:
                    queue.append((node_depth + 1, succ))

    def new_block_addr(self) -> int:
        """
        Return a block address that does not conflict with any existing blocks.

        :return:    The block address.
        """
        new_addr = max(self._new_block_addrs) + 1 if self._new_block_addrs else max(self.blocks_by_addr) + 2048
        self._new_block_addrs.add(new_addr)
        return new_addr

    def _get_block(self, addr, **kwargs) -> ailment.Block | None:
        """
        Get exactly one block by its address and optionally, also considering its block ID. An exception,
        MultipleBlocksException, will be raised if there are more than one block satisfying the specified criteria.

        :param addr:        The address of the block.
        :param kwargs:      Optionally, you can specify "idx" to consider the block ID. If "idx" is not specified, this
                            method will return the only block at the specified address, None if there is no block at
                            that address, or raise an exception if there are more than one block at that address.
        :return:            The requested block or None if no block matching the specified criteria exists.
        """

        if not self._blocks_by_addr:
            return None
        idx_specified = "idx" in kwargs
        idx = kwargs.get("idx")
        if not idx_specified:
            blocks = self._blocks_by_addr.get(addr, None)
        else:
            blocks = [self._blocks_by_addr_and_idx.get((addr, idx), None)]
        if not blocks:
            return None
        if len(blocks) == 1:
            return next(iter(blocks))
        if idx_specified:
            raise MultipleBlocksException(
                f"There are {len(blocks)} blocks at address {addr:#x}.{idx} but only one is requested."
            )
        raise MultipleBlocksException(
            f"There are {len(blocks)} blocks at address {addr:#x} (block ID ignored) but only one is requested."
        )

    def _get_blocks(self, addr, idx=None) -> Generator[ailment.Block]:
        if not self._blocks_by_addr:
            return
        else:
            if idx is None:
                blocks = self._blocks_by_addr.get(addr, None)
                if blocks is not None:
                    yield from blocks
            else:
                block = self._blocks_by_addr_and_idx.get((addr, idx), None)
                if block is not None:
                    yield block

    def _update_block(self, old_block, new_block):
        if self.out_graph is None:
            self.out_graph = self._graph  # we do not make copy here for performance reasons. we can change it if needed
            assert self.out_graph is not None

        if old_block not in self.out_graph:
            return

        in_edges = list(self.out_graph.in_edges(old_block, data=True))
        out_edges = list(self.out_graph.out_edges(old_block, data=True))

        self._remove_block(old_block)
        self.out_graph.add_node(new_block)
        self._blocks_by_addr[new_block.addr].add(new_block)
        self._blocks_by_addr_and_idx[(new_block.addr, new_block.idx)] = new_block

        for src, _, data in in_edges:
            if src is old_block:
                src = new_block
            self.out_graph.add_edge(src, new_block, **data)

        for _, dst, data in out_edges:
            if dst is old_block:
                dst = new_block
            self.out_graph.add_edge(new_block, dst, **data)

    def _remove_block(self, block):
        if self.out_graph is None:
            self.out_graph = self._graph
            assert self.out_graph is not None

        if block in self.out_graph:
            self.out_graph.remove_node(block)

        if block.addr in self._blocks_by_addr and block in self._blocks_by_addr[block.addr]:
            self._blocks_by_addr[block.addr].remove(block)
            del self._blocks_by_addr_and_idx[(block.addr, block.idx)]

    @staticmethod
    def _is_add(expr):
        return isinstance(expr, ailment.Expr.BinaryOp) and expr.op == "Add"

    @staticmethod
    def _is_sub(expr):
        return isinstance(expr, ailment.Expr.BinaryOp) and expr.op == "Sub"

    def _simplify_blocks(
        self,
        ail_graph: networkx.DiGraph,
        cache: dict | None = None,
    ):
        """
        Simplify all blocks in self._blocks.

        :param ail_graph:               The AIL function graph.
        :param cache:                   A block-level cache that stores reaching definition analysis results and
                                        propagation results.
        :return:                        None
        """

        blocks_by_addr_and_idx: dict[tuple[int, int | None], ailment.Block] = {}

        for ail_block in ail_graph.nodes():
            simplified = self._simplify_block(
                ail_block,
                cache=cache,
            )
            key = ail_block.addr, ail_block.idx
            blocks_by_addr_and_idx[key] = simplified

        # update blocks_map to allow node_addr to node lookup
        def _replace_node_handler(node):
            key = node.addr, node.idx
            if key in blocks_by_addr_and_idx:
                return blocks_by_addr_and_idx[key]
            return None

        AILGraphWalker(ail_graph, _replace_node_handler, replace_nodes=True).walk()

        return ail_graph

    def _simplify_block(self, ail_block, cache=None):
        """
        Simplify a single AIL block.

        :param ailment.Block ail_block: The AIL block to simplify.
        :return:                        A simplified AIL block.
        """

        cached_rd, cached_prop = None, None
        cache_item = None
        cache_key = ail_block.addr, ail_block.idx
        if cache:
            cache_item = cache.get(cache_key, None)
            if cache_item:
                # cache hit
                cached_rd = cache_item.rd
                cached_prop = cache_item.prop

        simp = self.project.analyses.AILBlockSimplifier(
            ail_block,
            self._func.addr,
            peephole_optimizations=self._peephole_optimizations,
            cached_reaching_definitions=cached_rd,
            cached_propagator=cached_prop,
        )
        # update the cache
        if cache is not None:
            if cache_item:
                del cache[cache_key]
            cache[cache_key] = BlockCache(simp._reaching_definitions, simp._propagator)
        return simp.result_block

    def _simplify_graph(self, graph):
        MAX_SIMP_ITERATION = 8
        for _ in range(MAX_SIMP_ITERATION):
            self._simplify_blocks(graph)
            simp = self.project.analyses.AILSimplifier(
                self._func,
                func_graph=graph,
                use_callee_saved_regs_at_return=False,
                gp=self._func.info.get("gp", None) if self.project.arch.name in {"MIPS32", "MIPS64"} else None,
                avoid_vvar_ids=self._avoid_vvar_ids,
            )
            if simp.simplified:
                graph = simp.func_graph
            else:
                break
        else:
            _l.warning("Failed to reach fixed point after %s simplification iterations.", MAX_SIMP_ITERATION)
        return graph

    def _recover_regions(self, graph: networkx.DiGraph, condition_processor=None, update_graph: bool = False):
        return self.project.analyses[RegionIdentifier].prep(kb=self.kb)(
            self._func,
            graph=graph,
            cond_proc=condition_processor or ConditionProcessor(self.project.arch),
            update_graph=update_graph,
            force_loop_single_exit=self._force_loop_single_exit,
            refine_loops_with_single_successor=self._refine_loops_with_single_successor,
            complete_successors=self._complete_successors,
            entry_node_addr=self.entry_node_addr,
        )


class SequenceOptimizationPass(BaseOptimizationPass):
    """
    The base class for any sequence node optimization pass.
    """

    def __init__(self, func, seq=None, **kwargs):
        super().__init__(func)
        self.seq = seq
        self.out_seq = None


class StructuringOptimizationPass(OptimizationPass):
    """
    The base class for any optimization pass that requires structuring. Optimization passes that inherit from this class
    should directly depend on structuring artifacts, such as regions and gotos. Otherwise, they should use
    OptimizationPass. This is the heaviest (computation time) optimization pass class.

    By default this type of optimization should work:
    - on any architecture
    - on any platform
    - during region identification (to have iterative structuring)
    - only with the SAILR structuring algorithm
    """

    ARCHES = None
    PLATFORMS = None
    STRUCTURING = [SAILRStructurer.NAME]
    STAGE = OptimizationPassStage.DURING_REGION_IDENTIFICATION

    _initial_gotos: set[Goto]
    _goto_manager: GotoManager | None
    _prev_graph: networkx.DiGraph

    def __init__(
        self,
        func,
        require_structurable_graph: bool = True,
        prevent_new_gotos: bool = True,
        strictly_less_gotos: bool = False,
        recover_structure_fails: bool = True,
        must_improve_rel_quality: bool = True,
        max_opt_iters: int = 1,
        simplify_ail: bool = True,
        require_gotos: bool = True,
        readd_labels: bool = False,
        edges_to_remove: list[tuple[tuple[int, int | None], tuple[int, int | None]]] | None = None,
        **kwargs,
    ):
        super().__init__(func, **kwargs)
        self._require_structurable_graph = require_structurable_graph
        self._prevent_new_gotos = prevent_new_gotos
        self._strictly_less_gotos = strictly_less_gotos
        self._recover_structure_fails = recover_structure_fails
        self._max_opt_iters = max_opt_iters
        self._simplify_ail = simplify_ail
        self._require_gotos = require_gotos
        self._must_improve_rel_quality = must_improve_rel_quality
        self._readd_labels = readd_labels
        self._edges_to_remove = edges_to_remove or []
        self._goto_manager = None
        self._initial_gotos = set()

        # relative quality metrics (excludes gotos)
        self._initial_structure_counter = None
        self._current_structure_counter = None

    def _analyze(self, cache=None) -> bool:
        raise NotImplementedError

    def analyze(self):
        """
        Wrapper for _analyze() that verifies the graph is structurable before and after the optimization.
        """
        # replace the normal check in OptimizationPass.analyze()
        ret, cache = self._check()
        if not ret:
            return

        # only initialize self._goto_manager if this optimization requires a structurable graph or gotos
        initial_structurable: bool | None = None
        if self._require_structurable_graph or self._require_gotos or self._prevent_new_gotos:
            initial_structurable = self._graph_is_structurable(self._graph, initial=True)

        if self._require_structurable_graph and initial_structurable is False:
            return

        if self._require_gotos:
            assert self._goto_manager is not None
            self._initial_gotos = self._goto_manager.gotos.copy()
            if not self._initial_gotos:
                return

        # setup for the very first analysis
        self.out_graph = networkx.DiGraph(self._graph)
        if self._max_opt_iters > 1:
            self._fixed_point_analyze(cache=cache)
        else:
            updates = self._analyze(cache=cache)
            if not updates:
                self.out_graph = None

        # analysis is complete, no out_graph means it failed somewhere along the way
        if self.out_graph is None:
            return

        # since all checks have completed, add labels back out here
        if self._readd_labels:
            self.out_graph = add_labels(self.out_graph)

        if (
            self._require_structurable_graph
            and self._max_opt_iters <= 1
            and not self._graph_is_structurable(self.out_graph, readd_labels=False)
        ):
            # fixed-point analysis ensures that the output graph is always structurable, otherwise it clears the output
            # graph. so we only check the structurability of the graph when fixed-point analysis did not run.
            self.out_graph = None
            return

        # simplify the AIL graph
        if self._simplify_ail:
            # this should not (TM) change the structure of the graph but is needed for later optimizations
            self.out_graph = self._simplify_graph(self.out_graph)

        if self._prevent_new_gotos:
            prev_gotos = len(self._initial_gotos)
            new_gotos = len(self._get_new_gotos())
            if (self._strictly_less_gotos and (new_gotos >= prev_gotos)) or (
                not self._strictly_less_gotos and (new_gotos > prev_gotos)
            ):
                self.out_graph = None
                return

        if self._must_improve_rel_quality and not self._improves_relative_quality():
            self.out_graph = None
            return

    def _get_new_gotos(self):
        assert self._goto_manager is not None
        return self._goto_manager.gotos

    def _fixed_point_analyze(self, cache=None):
        had_any_changes = False
        for _ in range(self._max_opt_iters):
            if self._require_gotos:
                assert self._goto_manager is not None
                if not self._goto_manager.gotos:
                    break

            # backup the graph before the optimization
            if self._recover_structure_fails and self.out_graph is not None:
                self._prev_graph = networkx.DiGraph(self.out_graph)

            # run the optimization, output applied to self.out_graph
            changes = self._analyze(cache=cache)
            if not changes:
                break

            had_any_changes = True
            # check if the graph is structurable
            if not self._graph_is_structurable(self.out_graph, readd_labels=self._readd_labels):
                if self._recover_structure_fails:
                    self.out_graph = self._prev_graph
                else:
                    self.out_graph = None
                    break

        if not had_any_changes:
            self.out_graph = None

    def _graph_is_structurable(self, graph, readd_labels=False, initial=False) -> bool:
        """
        Checks weather the input graph is structurable under the Phoenix schema-matching structuring algorithm.
        As a side effect, this will also update the region identifier and goto manager of this optimization pass.
        Consequently, a true return guarantees up-to-date goto information in the goto manager.
        """
        if readd_labels:
            graph = add_labels(graph)

        remove_edges_in_ailgraph(graph, self._edges_to_remove)

        self._ri = self.project.analyses[RegionIdentifier].prep(kb=self.kb)(
            self._func,
            graph=graph,
            # never update the graph in-place, we need to keep the original graph for later use
            update_graph=False,
            cond_proc=self._ri.cond_proc,
            force_loop_single_exit=False,
            complete_successors=True,
            entry_node_addr=self.entry_node_addr,
        )
        if self._ri is None:
            return False

        # we should try-catch structuring here because we can often pass completely invalid graphs
        # that break the assumptions of the structuring algorithm
        try:
            rs = self.project.analyses[RecursiveStructurer].prep(kb=self.kb)(
                self._ri.region,
                cond_proc=self._ri.cond_proc,
                func=self._func,
                structurer_cls=SAILRStructurer,
            )
        # pylint:disable=broad-except
        except Exception:
            _l.warning("Internal structuring failed for OptimizationPass on %s", self._func.name)
            rs = None

        if not rs or not rs.result or is_empty_node(rs.result) or rs.result_incomplete:
            return False

        rs = self.project.analyses.RegionSimplifier(self._func, rs.result, arg_vvars=self._arg_vvars, kb=self.kb)
        if not rs or rs.goto_manager is None or rs.result is None:
            return False

        self._analyze_simplified_region(rs.result, initial=initial)
        self._goto_manager = rs.goto_manager
        return True

    # pylint:disable=no-self-use
    def _analyze_simplified_region(self, region, initial=False):
        """
        Analyze the simplified regions after a successful structuring pass.
        This should be overridden by the subclass if it needs to do anything with the simplified regions for making
        optimizations decisions.
        """
        if region is None:
            return

        # record quality metrics
        if self._must_improve_rel_quality:
            if initial:
                self._initial_structure_counter = ControlFlowStructureCounter(region)
            else:
                self._current_structure_counter = ControlFlowStructureCounter(region)

    def _improves_relative_quality(self) -> bool:
        """
        Welcome to the unprincipled land of mahaloz. This function is a heuristic that tries to determine if the
        optimization pass improved the relative quality of the control flow structures in the function. These heuristics
        are based on mahaloz's observations of what bad code looks like.
        """
        if self._initial_structure_counter is None or self._current_structure_counter is None:
            _l.warning("Relative quality check failed due to missing structure counters")
            return True

        prev_wloops = self._initial_structure_counter.while_loops
        curr_wloops = self._current_structure_counter.while_loops
        prev_dloops = self._initial_structure_counter.do_while_loops
        curr_dloops = self._current_structure_counter.do_while_loops
        prev_floops = self._initial_structure_counter.for_loops
        curr_floops = self._current_structure_counter.for_loops
        total_prev_loops = prev_wloops + prev_dloops + prev_floops
        total_curr_loops = curr_wloops + curr_dloops + curr_floops

        # Sometimes, if we mess up structuring you can easily tell because we traded "good" loops for "bad" loops.
        # Generally, loops are ordered good -> bad as follows: for, while, do-while.
        # Note: this check is only for _trading_, meaning the total number of loops must be the same.
        #
        # 1. We traded to remove a for-loop
        if curr_floops < prev_floops and total_curr_loops == total_prev_loops:
            return False

        # Gotos play an important part in readability and control flow structure. We already count gotos in other parts
        # of the analysis, so we don't need to count them here. However, some gotos are worse than others. Much
        # like loops, trading gotos (keeping the same total, but getting worse types), is bad for decompilation.
        if self._goto_manager is not None and len(self._initial_gotos) == len(self._goto_manager.gotos) != 0:
            prev_labels = self._initial_structure_counter.goto_targets
            curr_labels = self._current_structure_counter.goto_targets

            # 1. We traded gotos, but we increased the number of labels, which is generally worse
            if len(curr_labels) > len(prev_labels):
                return False

            ordered_curr_labels = self._current_structure_counter.ordered_labels

            # 2. We trade for a goto that occurs higher in the program (much like a back edge goto), these are bad
            for addr, curr_cnt in curr_labels.items():
                prev_cnt = prev_labels.get(addr, 0)
                # some label increased in gotos, check everything to the right in ordered labels, if it went down,
                # then we fail
                if curr_cnt > prev_cnt:
                    right_labels = ordered_curr_labels[ordered_curr_labels.index(addr) + 1 :]
                    for right_label in right_labels:
                        right_curr_label_cnt = curr_labels[right_label]
                        right_prev_label_cnt = prev_labels.get(right_label, 0)
                        if right_curr_label_cnt < right_prev_label_cnt:
                            return False

                # some label decreased in gotos, check everything to the left, if something went up, then we fail
                elif curr_cnt < prev_cnt:
                    left_labels = ordered_curr_labels[: ordered_curr_labels.index(addr)]
                    for left_label in left_labels:
                        left_curr_label_cnt = curr_labels[left_label]
                        left_prev_label_cnt = prev_labels.get(left_label, 0)
                        if left_curr_label_cnt > left_prev_label_cnt:
                            return False

        return True



# Microsoft Visual C/C++ Runtime
from __future__ import annotations
from . import SimLibrary
from angr.procedures import SIM_PROCEDURES as P
from angr.calling_conventions import SimCCMicrosoftAMD64

libc = SimLibrary()
libc.set_library_names(
    "msvcrt.dll",
    "msvcr71.dll",
    "msvcr80.dll",
    "msvcr90.dll",
    "msvcr100.dll",
    "msvcr110.dll",
    "msvcrt20.dll",
    "msvcrt40.dll",
    "msvcr120.dll",
)
libc.add_all_from_dict(P["libc"])
libc.add_all_from_dict(P["msvcr"])  # overwrite any that are also defined in libc
libc.set_non_returning("_exit", "abort", "exit", "_invoke_watson")

libc.add_alias("_initterm", "_initterm_e")

libc.set_default_cc("AMD64", SimCCMicrosoftAMD64)
